import numpy
import pytz
import datetime
import dateutil.parser
import logging
import re

import astropy.io.fits as pyfits

from sourcefinder.utils import is_valid_beam_tuple
from sourcefinder.accessors.dataaccessor import DataAccessor
from sourcefinder.utility.coordinates import WCS

logger = logging.getLogger(__name__)


class FitsImage(DataAccessor):
    """Use PyFITS to pull image data out of a FITS file.

    Provide standard attributes, as per :class:`DataAccessor`. In addition, we
    provide a ``telescope`` attribute if the FITS file has a ``TELESCOP``
    header.

    Parameters
    ----------
    url : Path or str
        The path or URL to the FITS file.
    plane : int, default: None
        If the data is a datacube, specifies which plane to use.
    beam : tuple, default: None
        Beam parameters in degrees, in the form (bmaj, bmin, bpa). If not
        supplied, the method will attempt to read these from the header.
    hdu_index : int, default: 0
        The index of the HDU to use from the HDU list.

    """

    def __init__(self, url, plane=None, beam=None, hdu_index=0):
        self.url = url
        self.header = self._get_header(hdu_index)
        self.wcs = self.parse_coordinates()
        self.data = self.read_data(hdu_index, plane)
        self.taustart_ts, self.tau_time = self.parse_times()
        self.freq_eff, self.freq_bw = self.parse_frequency()
        self.pixelsize = self.parse_pixelsize()

        if is_valid_beam_tuple(beam) or not is_valid_beam_tuple(self.beam):
            # An argument-supplied beam overrides a beam derived from
            # (bmaj, bmin, bpa) in a config.toml. Only if those two options
            # are not specified, we parse the beam from the header.
            bmaj, bmin, bpa = (
                beam if is_valid_beam_tuple(beam) else (self.parse_beam())
            )
            self.beam = self.degrees2pixels(
                bmaj, bmin, bpa, self.pixelsize[0], self.pixelsize[1]
            )

        self.centre_ra, self.centre_decl = self.calculate_phase_centre()

        # Bonus attribute
        if "TELESCOP" in self.header:
            self.telescope = self.header["TELESCOP"]

    def _get_header(self, hdu_index):
        """Retrieve the header from the specified HDU in the FITS
        file.

        Parameters
        ----------
        hdu_index : int
            The index of the Header Data Unit (HDU) to extract the header from.

        Returns
        -------
        astropy.io.fits.Header
            A copy of the header from the specified HDU.

        """
        with pyfits.open(self.url) as hdulist:
            hdu = hdulist[hdu_index]
        return hdu.header.copy()

    def read_data(self, hdu_index, plane):
        """Read data from our FITS file.

        Parameters
        ----------
        hdu_index : int
            The index of the Header Data Unit (HDU) to extract the data from.
        plane : int, default: None
            If the data is a datacube, specifies which plane to use.

        Returns
        -------
        numpy.ndarray
            The processed 2D data array, transposed for intuitive display.

        Notes
        -----
        PyFITS reads the data into an array indexed as [y][x]. We
        take the transpose to make this more intuitively reasonable and
        consistent with (eg) ds9 display of the FitsFile. Transpose back
        before viewing the array with RO.DS9, saving to a FITS file,
        etc.

        """
        with pyfits.open(self.url) as hdulist:
            hdu = hdulist[hdu_index]
            data = numpy.float64(hdu.data.squeeze())
        if plane is not None and len(data.shape) > 2:
            data = data[plane].squeeze()
        n_dim = len(data.shape)
        if n_dim != 2:
            logger.warning(
                (
                    f"Loaded datacube with {n_dim:d} dimensions, "
                    "assuming Stokes I and taking plane 0."
                )
            )
            data = data[0, :, :]
        data = data.transpose()
        return data

    def parse_coordinates(self):
        """Parse header to return a WCS (World Coordinate System)
        object.

        Returns
        -------
        WCS
            A WCS object containing the coordinate system information
            extracted from the FITS file header.

        Raises
        ------
        TypeError
            If the coordinate system is not specified in the FITS header.

        Notes
        -----
        If units are not specified in the header, degrees are assumed by
        default.

        """
        header = self.header
        wcs = WCS()
        try:
            wcs.crval = header["crval1"], header["crval2"]
            wcs.crpix = header["crpix1"] - 1, header["crpix2"] - 1
            wcs.cdelt = header["cdelt1"], header["cdelt2"]
        except KeyError:
            msg = "Coordinate system not specified in FITS"
            logger.error(msg)
            raise TypeError(msg)
        try:
            wcs.ctype = header["ctype1"], header["ctype2"]
        except KeyError:
            wcs.ctype = "unknown", "unknown"
        try:
            wcs.crota = float(header["crota1"]), float(header["crota2"])
        except KeyError:
            wcs.crota = 0.0, 0.0
        try:
            wcs.cunit = header["cunit1"], header["cunit2"]
        except KeyError:
            # The "Definition of the Flexible Image Transport System", version
            # 3.0, tells us that "units for celestial coordinate systems defined
            # in this Standard must be degrees", so we assume that if nothing
            # else is specified.
            msg = "WCS units unknown; using degrees"
            logger.warning(msg)
            wcs.cunit = "deg", "deg"
        return wcs

    def calculate_phase_centre(self):
        """Calculate the phase center of the FITS image.

        The phase center is determined by finding the central pixel and
        converting that position to celestial coordinates.

        Returns
        -------
        tuple[float]
            A tuple containing the right ascension and declination of
            the phase center in degrees.

        """
        x, y = self.data.shape
        centre_ra, centre_decl = self.wcs.p2s((x / 2, y / 2))
        return float(centre_ra), float(centre_decl)

    def parse_frequency(self):
        """Set some 'shortcut' variables for access to the frequency
        parameters in the FITS file header.

        Returns
        -------
        freq_eff : float
            The effective frequency extracted from the FITS header.
        freq_bw : float
            The bandwidth extracted from the FITS header.

        """
        try:
            header = self.header
            if "RESTFRQ" in header:
                freq_eff = header["RESTFRQ"]
                if "RESTBW" in header:
                    freq_bw = header["RESTBW"]
                else:
                    logger.warning(
                        "bandwidth header missing in image {},"
                        " setting to 1 MHz".format(self.url)
                    )
                    freq_bw = 1e6
            elif ("CTYPE3" in header) and (
                header["CTYPE3"] in ("FREQ", "VOPT")
            ):
                freq_eff = header["CRVAL3"]
                freq_bw = header["CDELT3"]
            elif ("CTYPE4" in header) and (
                header["CTYPE4"] in ("FREQ", "VOPT")
            ):
                freq_eff = header["CRVAL4"]
                freq_bw = header["CDELT4"]
            else:
                freq_eff = header["RESTFREQ"]
                freq_bw = 1e6
        except KeyError:
            msg = "Frequency not specified in headers for {}".format(self.url)
            logger.error(msg)
            raise TypeError(msg)

        return freq_eff, freq_bw

    def parse_beam(self):
        """Read and return the beam properties bmaj, bmin and bpa
        values from the FITS header.

        Returns
        -------
        bmaj : float
            the major axis of the beam in degrees.
        bmin : float
            the minor axis of the beam in degrees.
        bpa : float
            the position angle of the beam in degrees.

        Notes
        -----
        AIPS FITS file: stored in the history section

        """
        beam_regex = re.compile(
            r"""
            BMAJ
            \s*=\s*
            (?P<bmaj>[-\d\.eE]+)
            \s*
            BMIN
            \s*=\s*
            (?P<bmin>[-\d\.eE]+)
            \s*
            BPA
            \s*=\s*
            (?P<bpa>[-\d\.eE]+)
            """,
            re.VERBOSE,
        )

        bmaj, bmin, bpa = None, None, None
        header = self.header
        try:
            # MIRIAD FITS file
            bmaj = header["BMAJ"]
            bmin = header["BMIN"]
            bpa = header["BPA"]
        except KeyError:

            def get_history(hdr):
                """Retrieve all history cards from a FITS header.

                Parameters
                ----------
                hdr : astropy.io.fits.Header
                    The FITS header object from which to extract history cards.

                Returns
                -------
                list[str]
                    A list of strings, where each string represents a
                    history card from the FITS header.

                """
                return hdr["HISTORY"]

            for hist_entry in get_history(header):
                results = beam_regex.search(hist_entry)
                if results:
                    bmaj, bmin, bpa = [
                        float(results.group(key))
                        for key in ("bmaj", "bmin", "bpa")
                    ]
                    break

        return bmaj, bmin, bpa

    def parse_times(self):
        """Attempt to do something sane with timestamps.

        Returns
        -------
        taustart_ts : datetime.datetime
            Timezone-naive (implicit UTC) datetime representing the
            start of the observation.
        tau_time : float
            Integration time in seconds.

        """
        try:
            start = self.parse_start_time()
        except KeyError:
            # If no start time specified, give up:
            logger.warning(
                (
                    "Timestamp not specified in FITS file:"
                    " using 'now' with dummy (zero-valued) integration "
                    "time."
                )
            )
            return datetime.datetime.now(), 0.0

        try:
            end = dateutil.parser.parse(self.header["end_utc"])
        except KeyError:
            msg = "End time not specified in {}, setting to start".format(
                self.url
            )
            logger.warning(msg)
            end = start

        delta = end - start
        tau_time = delta.total_seconds()

        # For simplicity, the database requires naive datetimes (implicit UTC)
        # So we convert to UTC and then drop the timezone:
        try:
            timezone = pytz.timezone(self.header["timesys"])
            start_w_tz = start.replace(tzinfo=timezone)
            start_utc = pytz.utc.normalize(start_w_tz.astimezone(pytz.utc))
            return start_utc.replace(tzinfo=None), tau_time
        except (pytz.UnknownTimeZoneError, KeyError):
            logger.debug("Timezone not specified in FITS file: assuming UTC.")
            return start, tau_time

    def parse_start_time(self):
        """Parse and return the start time of the observation, that
        yielded this FITS image, from its header.

        Returns
        -------
        datetime.datetime
            The start time of the observation as an instance of
            `datetime.datetime`.

        Raises
        ------
        KeyError
            If the timestamp in the FITS file is unreadable.
        Warning
            Logged if a non-standard date format is encountered in the FITS
            file.

        """
        header = self.header
        try:
            if ";" in header["date-obs"]:
                start = dateutil.parser.parse(
                    header["date-obs"].split(";")[0].split('"')[1]
                )
            else:
                start = dateutil.parser.parse(header["date-obs"])
        except AttributeError:
            # Maybe it's a float, Westerbork-style?
            if isinstance(header["date-obs"], float):
                logger.warning("Non-standard date specified in FITS file!")
                frac, year = numpy.modf(header["date-obs"])
                start = datetime.datetime(int(year), 1, 1)
                delta = datetime.timedelta(365.242199 * frac)
                start += delta
            else:
                raise KeyError("Timestamp in fits file unreadable")
        return start



from dataclasses import dataclass


@dataclass
class LofarAccessor():
    """
    Additional metadata required for processing LOFAR images through QC
    checks.

    Attributes
    ----------
    antenna_set : str
        Antenna set in use during observation. Possible values are
        'LBA_INNER', 'LBA_OUTER', 'LBA_SPARSE', 'LBA', or 'HBA'.
    ncore : int
        Number of core stations in use during observation.
    nremote : int
        Number of remote stations in use during observation.
    nintl : int
        Number of international stations in use during observation.
    subbandwidth : float
        Width of a subband in Hz.
    subbands : int
        Number of subbands.
    """

    antenna_set: str
    ncore: int
    nremote: int
    nintl: int
    subbandwidth: float
    subbands: int



import logging

import numpy

from sourcefinder.utils import is_valid_beam_tuple
from sourcefinder.accessors.fitsimage import FitsImage

logger = logging.getLogger(__name__)


class FitsImageBlob(FitsImage):
    """
    A FITS image blob. Same as ``sourcefinder.accessors.fitsimage.FitsImage``
    but constructed from an in-memory FITS file, not a FITS file on disk.

    Parameters
    ----------
    hdulist : astropy.io.fits.HDUList
        The HDU list representing the in-memory FITS file.
    plane : int, default: None
        If the data is a datacube, specifies which plane to use.
    beam : tuple, default: None
        Beam parameters in degrees, in the form (bmaj, bmin, bpa). If not 
        supplied, the method will attempt to read these from the header.
    hdu_index : int, default: 0
        The index of the HDU to use from the HDU list.
    """
    def __init__(self, hdulist, plane=None, beam=None, hdu_index=0):
        # Set the URL in case we need it during header parsing for error
        # logging.
        self.url = "AARTFAAC streaming image"

        self.header = self._get_header(hdulist, hdu_index)
        self.wcs = self.parse_coordinates()
        self.data = self.read_data(hdulist, hdu_index, plane)
        self.taustart_ts, self.tau_time = self.parse_times()
        self.freq_eff, self.freq_bw = self.parse_frequency()
        self.pixelsize = self.parse_pixelsize()

        elements = "memory://AARTFAAC", self.taustart_ts, self.tau_time, \
                   self.freq_eff, self.freq_bw
        self.url = "_".join([str(x) for x in elements])

        if is_valid_beam_tuple(beam) or not is_valid_beam_tuple(self.beam):
            # An argument-supplied beam overrides a beam derived from
            # (bmaj, bmin, bpa) in a config.toml. Only if those two options
            # are not specified, we parse the beam from the header.
            bmaj, bmin, bpa = beam if is_valid_beam_tuple(beam) else (
                self.parse_beam())
            self.beam = self.degrees2pixels(
                bmaj, bmin, bpa, self.pixelsize[0], self.pixelsize[1]
            )

        self.centre_ra, self.centre_decl = self.calculate_phase_centre()

        # Bonus attribute
        if 'TELESCOP' in self.header:
            self.telescope = self.header['TELESCOP']

    def _get_header(self, *args):
        """
        Retrieve the header from the specified HDU.

        Parameters
        ----------
        *args : tuple
            Positional arguments where:

            - args[0] is the HDU list (astropy.io.fits.HDUList).
            - args[1] is the index of the HDU to use.

        Returns
        -------
        astropy.io.fits.Header
            The header of the specified HDU.
        """
        hdulist = args[0]
        hdu_index = args[1]
        return hdulist[hdu_index].header

    def read_data(self, *args):
        """
        Read and process the data from the specified HDU.

        Parameters
        ----------
        *args : tuple
            Positional arguments where:

            - args[0] is the HDU list (astropy.io.fits.HDUList).
            - args[1] is the index of the HDU to use.
            - args[2] is the plane index (int) if the data is a datacube.

        Returns
        -------
        numpy.ndarray
            The processed 2D data array. Processing here means remove axes of
            length 1, select the plane index from the datacube if needed, and
            transpose.
        """
        hdulist = args[0]
        hdu_index = args[1]
        plane = args[2]
        hdu = hdulist[hdu_index]
        data = numpy.float64(hdu.data.squeeze())
        if plane is not None and len(data.shape) > 2:
            data = data[plane].squeeze()
        n_dim = len(data.shape)
        if n_dim != 2:
            logger.warning((
                "Loaded datacube with %s dimensions, assuming Stokes I and "
                "taking plane 0" % n_dim))
            data = data[0, :, :]
        data = data.transpose()
        return data



"""
This module implements the CASA LOFAR data container format, described in this
document:

http://www.lofar.org/operations/lib/exe/fetch.php?media=:public:documents
:casa_image_for_lofar_0.03.00.pdf
"""
import datetime
import logging

import numpy
from casacore.tables import table as casacore_table

from sourcefinder.accessors.casaimage import CasaImage
from sourcefinder.accessors.lofaraccessor import LofarAccessor
from sourcefinder.utility.coordinates import julian2unix

logger = logging.getLogger(__name__)

subtable_names = (
    'LOFAR_FIELD',
    'LOFAR_ANTENNA',
    'LOFAR_HISTORY',
    'LOFAR_ORIGIN',
    'LOFAR_QUALITY',
    'LOFAR_STATION',
    'LOFAR_POINTING',
    'LOFAR_OBSERVATION'
)


class LofarCasaImage(CasaImage, LofarAccessor):  # type: ignore[misc]
    """
    Use casacore to pull image data out of a CASA table.

    This accessor assumes the casatable contains the values described in the
    CASA Image description for LOFAR. 0.03.00.

    Parameters
    ----------
    url : str
        Location of the CASA table.
    plane : int, default: 0
        If the data is a datacube, specifies which plane to use.
    beam : tuple, default: None
        Beam parameters in degrees, in the form (bmaj, bmin, bpa). If not
        supplied, the method will attempt to read these from the header.
    """

    def __init__(self, url, plane=0, beam=None):
        super().__init__(url, plane, beam)
        table = casacore_table(self.url, ack=False)
        subtables = self.open_subtables(table)
        self.taustart_ts = self.parse_taustartts(subtables)
        self.tau_time = self.parse_tautime(subtables)

        # Additional, LOFAR-specific metadata
        self.antenna_set = self.parse_antennaset(subtables)
        self.ncore, self.nremote, self.nintl = self.parse_stations(subtables)
        self.subbandwidth = self.parse_subbandwidth(subtables)
        self.subbands = self.parse_subbands(subtables)

    @staticmethod
    def open_subtables(table):
        """Open all subtables defined in the LOFAR format.

        Parameters
        ----------
        table : casacore.tables.table
            A casacore table handler to a LOFAR CASA table.

        Returns
        -------
        dict
            A dictionary containing all LOFAR CASA subtables

        """
        subtables = {}
        for subtable in subtable_names:
            subtable_location = table.getkeyword("ATTRGROUPS")[subtable]
            subtables[subtable] = casacore_table(subtable_location, ack=False)
        return subtables

    @staticmethod
    def parse_taustartts(subtables):
        """Extract the image start time from the CASA table header.
        
        Parameters
        ----------
        subtables : dict
            A dictionary containing all LOFAR CASA subtables.
        
        Returns
        -------
        datetime.datetime
            The earliest observation start time as a datetime object.
        
        Notes
        -----
        We sort the table in order of ascending start time then
        choose the first value to ensure we get the earliest possible
        starting time.

        """
        observation_table = subtables['LOFAR_OBSERVATION']
        julianstart = observation_table.query(
            sortlist="OBSERVATION_START", limit=1).getcell(
            "OBSERVATION_START", 0
        )
        unixstart = julian2unix(julianstart)
        taustart_ts = datetime.datetime.fromtimestamp(unixstart)
        return taustart_ts

    @staticmethod
    def non_overlapping_time(series):
        """Calculate the sum of total time ranges without overlap.

        Parameters
        ----------
        series : list of tuple
            A list of 2-item tuples representing time ranges, where each tuple
            contains the start and end time of a range.

        Returns
        -------
        float
            The total length of all time ranges without overlap, most likely in
            seconds.

        """
        series.sort()
        overlap = total = 0
        for n, (start, end) in enumerate(series):
            total += end - start
            for (nextstart, nextend) in series[n + 1:]:
                if nextstart >= end:
                    break
                overlapstart = max(nextstart, start)
                overlapend = min(nextend, end)
                overlap += overlapend - overlapstart
                start = overlapend
        return total - overlap

    @staticmethod
    def parse_tautime(subtables):
        """Returns the total on-sky time for this image.
        
        Parameters
        ----------
        subtables : dict
            A dictionary containing all LOFAR CASA subtables.
        
        Returns
        -------
        float
            The total on-sky time, most likely in seconds.

        """
        origin_table = subtables['LOFAR_ORIGIN']
        startcol = origin_table.col('START')
        endcol = origin_table.col('END')
        series = [(int(start), int(end)) for start, end in
                  zip(startcol, endcol)]
        tau_time = LofarCasaImage.non_overlapping_time(series)
        return tau_time

    @staticmethod
    def parse_antennaset(subtables):
        """Extract the antenna set used in the observation.

        This method retrieves the unique antenna set from the
        LOFAR_OBSERVATION subtable.

        Parameters
        ----------
        subtables : dict
            A dictionary containing all LOFAR CASA subtables.

        Returns
        -------
        numpy.ndarray
            An array containing the unique antenna set used in the observation.

        Raises
        ------
        Exception if multiple antenna sets are found.

        Notes
        -----
        The method uses the `unique_column_values` function from the
        `CasaImage` class to extract unique values from the "ANTENNA_SET"
        column.

        """
        observation_table = subtables['LOFAR_OBSERVATION']
        antennasets = CasaImage.unique_column_values(observation_table,
                                                     "ANTENNA_SET")
        if len(antennasets) == 1:
            return antennasets[0]
        else:
            raise Exception("Cannot handle multiple antenna sets in image")

    @staticmethod
    def parse_subbands(subtables):
        """Extract the number of subbands used in the observation.

        This method retrieves the unique number of subbands from the
        LOFAR_ORIGIN subtable. If multiple values are found, an exception is
        raised.

        Parameters
        ----------
        subtables : dict
            A dictionary containing all LOFAR CASA subtables.

        Returns
        -------
        int
            The number of subbands used in the observation.

        Raises
        ------
        Exception
            If varying numbers of channels are found in the subtable.

        Notes
        -----
        The method uses the `unique_column_values` function from the
        `CasaImage` class to extract unique values from the "NUM_CHAN" column.

        """
        origin_table = subtables['LOFAR_ORIGIN']
        num_chans = CasaImage.unique_column_values(origin_table, "NUM_CHAN")
        if len(num_chans) == 1:
            return num_chans[0]
        else:
            raise Exception(
                "Cannot handle varying numbers of channels in image")

    @staticmethod
    def parse_subbandwidth(subtables):
        """Calculate the subband width for the observation.

        This method determines the subband width based on the clock frequency
        retrieved from the LOFAR_OBSERVATION subtable.

        Parameters
        ----------
        subtables : dict
            A dictionary containing all LOFAR CASA subtables.

        Returns
        -------
        float
            The subband width in Hz.

        Raises
        ------
        Exception
            If multiple clock frequencies are found in the subtable.

        Notes
        -----
        The method uses the clock frequency and its unit to calculate the
        subband width. The clock frequency is divided by 1024 to determine
        the subband width. For more details, see:
        https://www.astron.nl/lofarwiki/doku.php?id=public:documents:
        raw_olap_data_formats&s[]=subband
        'base_subband_hz = clock_hz / 1024'

        """
        freq_units = {
            'Hz': 1,
            'kHz': 10 ** 3,
            'MHz': 10 ** 6,
            'GHz': 10 ** 9,
        }
        observation_table = subtables['LOFAR_OBSERVATION']
        clockcol = observation_table.col('CLOCK_FREQUENCY')
        clock_values = CasaImage.unique_column_values(observation_table,
                                                      "CLOCK_FREQUENCY")
        if len(clock_values) == 1:
            clock = clock_values[0]
            unit = clockcol.getkeyword('QuantumUnits')[0]
            trueclock = freq_units[unit] * clock
            subbandwidth = trueclock / 1024
            return subbandwidth
        else:
            raise Exception("Cannot handle varying clocks in image")

    @staticmethod
    def parse_stations(subtables):
        """Extract the number of specific LOFAR stations used in the
        observation.

        This method calculates the number of core, remote, and international
        LOFAR stations used based on the observation and antenna subtables.

        Parameters
        ----------
        subtables : dict
            A dictionary containing all LOFAR CASA subtables.

        Returns
        -------
        tuple
            A tuple containing the number of core stations, remote stations,
            and international stations in the format (ncore, nremote, nintl).

        Notes
        -----
        The method uses the "NVIS_USED" column from the LOFAR_OBSERVATION
        subtable and the "NAME" column from the LOFAR_ANTENNA subtable to
        determine which
        stations were used. Stations are categorized based on their names:

        - Core stations start with "CS".
        - Remote stations start with "RS".
        - International stations have other prefixes.

        """
        observation_table = subtables['LOFAR_OBSERVATION']
        antenna_table = subtables['LOFAR_ANTENNA']
        nvis_used = observation_table.getcol('NVIS_USED')
        names = numpy.array(antenna_table.getcol('NAME'))
        mask = numpy.sum(nvis_used, axis=2) > 0
        used = names[mask[0]]
        ncore = nremote = nintl = 0
        for station in used:
            if station.startswith('CS'):
                ncore += 1
            elif station.startswith('RS'):
                nremote += 1
            else:
                nintl += 1
        return ncore, nremote, nintl



"""Gaussian deconvolution."""

from math import sin, cos, atan, sqrt, pi
from numba import njit


@njit
def deconv(fmaj, fmin, fpa, cmaj, cmin, cpa):
    """Deconvolve a Gaussian "beam" from a Gaussian component.

    When we fit an elliptical Gaussian to a point in our image, we are
    actually fitting to a convolution of the physical shape of the source with
    the beam pattern of our instrument. This results in the fmaj/fmin/fpa
    arguments to this function.

    Since the shape of the (clean) beam (arguments cmaj/cmin/cpa) is known, we
    can deconvolve it from the fitted parameters to get the "real" underlying
    physical source shape, which is what this function returns.

    Parameters
    ----------
    fmaj : float
        Fitted major axis (pixels).
    fmin : float
        Fitted minor axis (pixels).
    fpa : float
        Fitted position angle of the major axis (degrees).
    cmaj : float
        Clean beam major axis (pixels).
    cmin : float
        Clean beam minor axis (pixels).
    cpa : float
        Clean beam position angle of the major axis (degrees).

    Returns
    -------
    rmaj : float
        real major axis in pixels
    rmin : float
        real minor axis in pixels
    rpa : float
        real position angle of the major axis in degress
    ierr : int
        number of components which failed to deconvolve

    Notes
    -----
    Instead of fmaj, fmin, cmaj and cmin all in pixels, one could use any
    arbitrary unit of sky angular distance, such as arcseconds or radians.
    The first two elements of the returned tuple would then have that same
    unit, while the third element (the position angle) would still be in
    degrees.

    """
    HALF_RAD = 90.0 / pi
    cmaj2 = cmaj * cmaj
    cmin2 = cmin * cmin
    fmaj2 = fmaj * fmaj
    fmin2 = fmin * fmin
    theta = (fpa - cpa) / HALF_RAD
    det = ((fmaj2 + fmin2) - (cmaj2 + cmin2)) / 2.0
    rhoc = (fmaj2 - fmin2) * cos(theta) - (cmaj2 - cmin2)
    sigic2 = 0.0
    rhoa = 0.0
    ierr = 0

    if abs(rhoc) > 0.0:
        sigic2 = atan((fmaj2 - fmin2) * sin(theta) / rhoc)
        rhoa = ((cmaj2 - cmin2) - (fmaj2 - fmin2) * cos(theta)) / (
            2.0 * cos(sigic2)
        )

    rpa = sigic2 * HALF_RAD + cpa
    rmaj = det - rhoa
    rmin = det + rhoa

    if rmaj < 0:
        ierr += 1
        rmaj = 0
    if rmin < 0:
        ierr += 1
        rmin = 0

    rmaj = sqrt(rmaj)
    rmin = sqrt(rmin)
    if rmaj < rmin:
        rmaj, rmin = rmin, rmaj
        rpa += 90

    rpa = (rpa + 900) % 180
    if not abs(rmaj):
        rpa = 0.0
    elif not abs(rmin) and (45.0 < abs(rpa - fpa) < 135.0):
        rpa = (rpa + 450.0) % 180.0

    return rmaj, rmin, rpa, ierr



"""Some generic utility routines for number handling and calculating
(specific) variances

"""

import itertools
import logging

import numpy as np
from numba import guvectorize, float32, int32

from sourcefinder import extract
from sourcefinder import stats
from sourcefinder import utils
from sourcefinder.config import Conf, ImgConf, ExportSettings
from sourcefinder.utility import containers
from sourcefinder.utility.uncertain import Uncertain
from sourcefinder.utility.sourceparams import make_measurements_dataframe
from concurrent.futures import ThreadPoolExecutor
from functools import cached_property
from functools import partial

from scipy import ndimage
from numba import guvectorize, float32, int32
import os

logger = logging.getLogger(__name__)


class ImageData(object):
    """Encapsulates an image in terms of a numpy array + meta/headerdata.

    This is your primary contact point for interaction with images: it
    includes facilities for source extraction and measurement, etc.

    Parameters
    ----------
    data : 2D np.ndarray
        Observational image data. Must be a regular np.ndarray, since image
        data read from e.g. a FITS file is not a MaskedArray.
    beam : tuple
        Clean beam specification as (semi-major axis, semi-minor axis,
        position angle) with the axes in pixel coordinates and the position
        angle in radians
    wcs : utility.coordinates.wcs
        World coordinate system specification, in our case it is always
        about sky coordinates.
    conf : Conf, default: Conf(image=ImgConf(), export=ExportSettings())
        Configuration options for source finding. This includes settings
        related to image processing (e.g., background and rms
        noise estimation, thresholds) as well as export options (e.g., source
        parameters and output maps).

    """

    def __init__(
        self,
        data,
        beam,
        wcs,
        conf: Conf = Conf(image=ImgConf(), export=ExportSettings()),
    ):
        # Do data, wcs and beam need deepcopy?
        # Probably not (memory overhead, in particular for data),
        # but then the user shouldn't change them outside ImageData in the
        # meantime
        # self.rawdata is a 2D numpy array.
        # Single precision is good enough in all cases.
        self.rawdata = np.ascontiguousarray(data, dtype=np.float32)
        self.wcs = wcs  # a utility.coordinates.wcs instance

        if utils.is_valid_beam_tuple(beam):
            self.beam = beam  # tuple of (semimaj, semimin, theta) with
            # semimaj and semimin in pixel coordinates and theta, the position
            # angle, in radians.
        else:
            raise ValueError(
                (
                    "Partial beam specification: one or more of "
                    "(bmaj, bmin, bpa) are not specified, "
                    "adequately, image processing is not possible.",
                    RuntimeWarning,
                )
            )

        # These three quantities are only dependent on the beam, so should be
        # calculated once the beam is known and not for each source separately.
        self.fudge_max_pix_factor = utils.fudge_max_pix(
            beam[0], beam[1], beam[2]
        )
        self.beamsize = utils.calculate_beamsize(beam[0], beam[1])
        self.correlation_lengths = utils.calculate_correlation_lengths(
            beam[0], beam[1]
        )
        self.clip: dict[float, np.ndarray] = {}
        self.labels: dict[float, tuple[np.ndarray, int]] = {}
        self._conf = conf

    @property
    def conf(self) -> Conf:
        return self._conf

    ###########################################################################
    #                                                                         #
    # Properties and attributes.                                              #
    #                                                                         #
    # Properties are attributes managed by methods; rather than calling the   #
    # method directly, the attribute automatically invokes it. Result of the  #
    # function calls are cached, subsequent calls doesn't recompute.          #
    #                                                                         #
    # clearcache() clears all the cached data, which can get quite large.     #
    # It may be wise to call this, for example, in an exception handler       #
    # dealing with MemoryErrors.                                              #
    #                                                                         #
    ###########################################################################
    @cached_property
    def grids(self):
        """Gridded RMS and background data for interpolating"""
        return self.__grids()

    @cached_property
    def backmap(self):
        """Mean background map"""
        if not hasattr(self, "_user_backmap"):
            return self._interpolate(self.grids["mean"], self.grids["indices"])
        else:
            return self._user_backmap

    @cached_property
    def rmsmap(self):
        """root-mean-squares map, i.e. the standard deviation of the local
        background noise, interpolated across the image."""
        if not hasattr(self, "_user_noisemap"):
            return self._interpolate(
                self.grids["rms"], self.grids["indices"], roundup=True
            )
        else:
            return self._user_noisemap

    @cached_property
    def data(self):
        """Masked image data"""
        # We will ignore all the data which is masked for the rest of the
        # sourcefinding process. We build up the mask by stacking ("or-ing
        # together") a number of different effects:
        #
        # * A margin from the edge of the image;
        # * Any data outside a given radius from the centre of the image;
        # * Data which is "obviously" bad (NaN).
        mask = np.zeros((self.xdim, self.ydim))
        if self.conf.image.margin:
            margin_mask = np.ones((self.xdim, self.ydim))
            margin_mask[
                self.conf.image.margin : -self.conf.image.margin,
                self.conf.image.margin : -self.conf.image.margin,
            ] = 0
            mask = np.logical_or(mask, margin_mask)
        if self.conf.image.radius:
            radius_mask = utils.circular_mask(
                self.xdim, self.ydim, self.conf.image.radius
            )
            mask = np.logical_or(mask, radius_mask)
        mask = np.logical_or(mask, np.isnan(self.rawdata))
        return np.ma.array(self.rawdata, mask=mask)

    @cached_property
    def data_bgsubbed(self):
        """Background subtracted masked image data"""
        return (self.data - self.backmap).astype(np.float32, copy=False)

    @property
    def xdim(self):
        """X pixel dimension of (unmasked) data"""
        return self.rawdata.shape[0]

    @property
    def ydim(self):
        """Y pixel dimension of (unmasked) data"""
        return self.rawdata.shape[1]

    @property
    def pixmax(self):
        """Maximum pixel value (pre-background subtraction)"""
        return self.data.max()

    @property
    def pixmin(self):
        """Minimum pixel value (pre-background subtraction)"""
        return self.data.min()

    def clearcache(self):
        """Zap any calculated data stored in this object.

        Clear the background and rms maps, labels, clip, and any locally held
        data. All of these can be reconstructed from the data accessor.

        Note that this *must* be run to pick up any new settings.

        """
        try:
            self.labels.clear()
            self.clip.clear()
            del self.backmap
            del self.rmsmap
            del self.data
            del self.data_bgsubbed
            del self.grids
            del self.Gaussian_islands
            del self.Gaussian_residuals
            del self.residuals_from_deblending
        except AttributeError:
            pass

    ###########################################################################
    #                                                                         #
    # General purpose image handling.                                         #
    #                                                                         #
    # Routines for saving and trimming data, and calculating background/RMS   #
    # maps (in conjuntion with the properties above).                         #
    #                                                                         #
    ###########################################################################

    # Private "support" methods
    def __grids(self):
        """Calculate background and RMS grids of this image.

        These grids can be interpolated up to make maps of the original image
        dimensions: see _interpolate().

        This is called automatically when ImageData.backmap,
        ImageData.rmsmap or ImageData.fdrmap is first accessed.

        """

        # there's no point in working with the whole of the data array
        # if it's masked.
        useful_chunk = ndimage.find_objects(np.where(self.data.mask, 0, 1))
        assert len(useful_chunk) == 1
        x_dim, y_dim = self.data[useful_chunk[0]].shape
        # Use 'back-size-x' and 'back-size-y' if available, fall back to 'grid'.
        back_size_x = self.conf.image.back_size_x or self.conf.image.grid
        if back_size_x is None:
            raise ValueError(
                (
                    "Expected either back-size-x or grid to be set "
                    "in the config object"
                )
            )
        back_size_y = self.conf.image.back_size_y or self.conf.image.grid
        if back_size_y is None:
            raise ValueError(
                (
                    "Expected either back-size-y or grid to be set "
                    "in the config object"
                )
            )
        # We should divide up the image into subimages such that each grid
        # node is centered on a subimage. This is only possible if
        # self.back_size_x and self.back_size_y are divisors of xdim and ydim,
        # respectively. If not, we need to select a frame within useful_chunk
        # that does have the appropriate dimensions. At the same time, it
        # should be as large as possible and centered within useful_chunk.
        rem_row = np.mod(x_dim, back_size_x)
        rem_col = np.mod(y_dim, back_size_y)

        start_offset_row, rem_rem_row = divmod(rem_row, 2)
        start_offset_col, rem_rem_col = divmod(rem_col, 2)
        end_offset_row = start_offset_row + rem_rem_row
        end_offset_col = start_offset_col + rem_rem_col

        offsets = np.array(
            [
                start_offset_row,
                -end_offset_row,
                start_offset_col,
                -end_offset_col,
            ]
        )

        useful_chunk_inds = ImageData.slices_to_indices(useful_chunk)[0]

        centred_inds = useful_chunk_inds + offsets

        # Before proceeding, check that our data has the size of at least
        # one subimage, for both dimensions.
        if (
            centred_inds[1] - centred_inds[0] > back_size_x
            and centred_inds[3] - centred_inds[2] > back_size_y
        ):

            subimages, number_of_elements_for_each_subimage = (
                utils.make_subimages(
                    self.data.data[
                        centred_inds[0] : centred_inds[1],
                        centred_inds[2] : centred_inds[3],
                    ],
                    self.data.mask[
                        centred_inds[0] : centred_inds[1],
                        centred_inds[2] : centred_inds[3],
                    ],
                    back_size_x,
                    back_size_y,
                )
            )

            mean_grid = np.zeros(
                number_of_elements_for_each_subimage.shape, dtype=np.float32
            )
            rms_grid = np.zeros(
                number_of_elements_for_each_subimage.shape, dtype=np.float32
            )

            stats.data_clipper_dynamic(
                subimages,
                number_of_elements_for_each_subimage,
                mean_grid,
                rms_grid,
            )

            # Fill in the zeroes with nearest neighbours.
            # In this way we do not have to make a MaskedArray, which
            # scipy.interpolate.interp1d cannot handle adequately.
            mean_grid = utils.nearest_nonzero(mean_grid, rms_grid)
            rms_grid = utils.nearest_nonzero(rms_grid, rms_grid)
        else:
            # Return an empty grid if we don't have enough pixels along both
            # dimensions. In that case ImageData._interpolate will return
            # completely masked background maps, which is what we want, since
            # no sources will be extracted.
            mean_grid = np.empty((0, 0), dtype=np.float32)
            rms_grid = np.empty((0, 0), dtype=np.float32)

        return {"mean": mean_grid, "rms": rms_grid, "indices": centred_inds}

    def _interpolate(self, grid, inds, roundup=False):
        """Interpolate a grid to produce a map of the dimensions of
        the image.


        Parameters
        ----------
        grid : np.ma.MaskedArray
            The grid to be interpolated.

        roundup : bool, default: False
            If True, values of the resultant map which are lower than the input
            grid are trimmed. Default is False.

        Returns
        -------
        np.ma.MaskedArray
            The interpolated map.

        Notes
        -----
        This function is used to transform the RMS, background or FDR grids
        produced by :func:`_grids()` to a map we can compare with the image
        data.

        """
        # Use zeroes with the mask from the observational image as a starting
        # point for the mean background and rms background maps. Next, use
        # the interpolated values from the background grids, which were derived
        # using kappa * sigma clipping, to fill in all unmasked pixels.
        my_map = np.ma.MaskedArray(
            np.zeros(self.data.shape), mask=self.data.mask, dtype=np.float32
        )

        # Remove the MaskedArrayFutureWarning warning and keep old numpy < 1.11
        # behavior
        my_map.unshare_mask()

        # If the grid has size 0 there is no point in proceeding.
        if grid.size == 0:
            my_map.mask = True
            return my_map

        my_xdim, my_ydim = inds[1] - inds[0], inds[3] - inds[2]

        if self.conf.image.median_filter:
            f_grid = ndimage.median_filter(grid, self.conf.image.median_filter)
            if self.conf.image.mf_threshold:
                grid = np.where(
                    np.fabs(f_grid - grid) > self.conf.image.mf_threshold,
                    f_grid,
                    grid,
                )
            else:
                grid = f_grid

        if self.conf.image.interpolate_order == 1:
            my_map[inds[0] : inds[1], inds[2] : inds[3]] = (
                utils.two_step_interp(grid, my_xdim, my_ydim)
            )
        else:
            # This condition is there to make sure we actually have some
            # unmasked patch of the image to fill.
            slicex = slice(-0.5, -0.5 + grid.shape[0], 1j * my_xdim)
            slicey = slice(-0.5, -0.5 + grid.shape[1], 1j * my_ydim)

            # Bicubic spline interpolation
            my_map[inds[0] : inds[1], inds[2] : inds[3]] = (
                ndimage.map_coordinates(
                    grid,
                    np.mgrid[slicex, slicey],
                    mode="nearest",
                    order=self.conf.image.interpolate_order,
                )
            )

        # If the input grid was entirely masked, then the output map must
        # also be masked: there's no useful data here. We don't search for
        # sources on a masked background/RMS, so this data will be cleanly
        # skipped by the rest of the sourcefinder
        if np.ma.getmask(grid).all():
            my_map.mask = True
        elif roundup:
            # In some cases, the spline interpolation may produce values
            # lower than the minimum value in the map. If required, these
            # can be trimmed off. No point doing this if the map is already
            # fully masked, though.
            my_map = np.ma.MaskedArray(
                data=np.where(my_map >= np.min(grid), my_map, np.min(grid)),
                mask=my_map.mask,
            )
        return my_map

    ###########################################################################
    #                                                                         #
    # Source extraction.                                                      #
    #                                                                         #
    # Provides for both traditional (islands-above-RMS) and FDR source        #
    # extraction systems.                                                     #
    #                                                                         #
    ###########################################################################

    def extract(
        self,
        noisemap=None,
        bgmap=None,
        labelled_data=None,
        labels=None,
    ):
        """Kick off conventional (ie, rms island finding) source
        extraction.

        Parameters
        ----------
        noisemap : np.ndarray, default: None
            Noise map, i.e. the standard deviation (rms) of the background
            noise across the observational image
        bgmap : np.ndarray, default: None
            Background map, i.e. the mean of the background noise across
            the observational image.
        labelled_data : np.ndarray, default: None
            The output of a connected component
            analysis of the image, with a unique label for each source. Should
            have the same shape as the observational image.
        labels : np.ndarray, default: None
            Labels array, i.e. a 1D integer array of labels for each source.

        Returns
        -------
        A `utility.containers.ExtractionResults` instance or a
        Pandas DataFrame containing the results of the source
        extraction.

        """
        if self.conf.image.analysis_thr > self.conf.image.detection_thr:
            logger.warning(
                "Analysis threshold is higher than detection threshold"
            )

        # If the image data is flat we may as well crash out here with a
        # sensible error message, otherwise the RMS estimation code will
        # crash out with a confusing error later.
        if np.ma.max(self.data) == np.ma.min(self.data):
            raise RuntimeError("Bad data: Image data is flat")

        if isinstance(bgmap, (np.ndarray, np.ma.MaskedArray)):
            if bgmap.shape != self.data.shape:
                raise IndexError("Background map has wrong shape")
            else:
                self.backmap = bgmap

        if isinstance(noisemap, (np.ndarray, np.ma.MaskedArray)):
            if noisemap.shape != self.data.shape:
                raise IndexError("Noisemap has wrong shape")
            if noisemap.min() < 0:
                raise ValueError("RMS noise cannot be negative")
            else:
                self.rmsmap = noisemap

        if (
            labelled_data is not None
            and labelled_data.shape != self.data.shape
        ):
            raise ValueError("Labelled map is wrong shape")

        return self._pyse(
            self.conf.image.detection_thr * self.rmsmap,
            self.conf.image.analysis_thr * self.rmsmap,
            labelled_data=labelled_data,
            labels=labels,
        )

    def reverse_se(self):
        """Run source extraction on the negative of this image.

        This process can be used to estimate the false positive rate, as there
        should be no sources in the negative image.

        Returns
        -------
        :class:`sourcefinder.utility.containers.ExtractionResults`

        To prevent interference with the normal extraction process, cached
        data (background map, clips, etc.) is cleared before and after
        running this method. If this method is used frequently, a separate
        cache may be implemented in the future.

        """
        self.labels.clear()
        self.clip.clear()
        self.data_bgsubbed *= -1
        results = self.extract()
        self.data_bgsubbed *= -1
        self.labels.clear()
        self.clip.clear()
        return results

    def fd_extract(
        self,
        alpha,
        noisemap=None,
        bgmap=None,
    ):
        """False Detection Rate based source extraction.

        The FDR procedure guarantees that the False Detection Rate (FDR) is
        less than alpha.

        Parameters
        ----------
        alpha : float
            Maximum allowed fraction of false positives. Must be between 0 and
            1, exclusive.
        noisemap : np.ndarray, default: None
            Noise map, i.e. the standard deviation (rms) of the background
            noise across the observational image
        bgmap : np.ndarray, default: None
            Background map, i.e. the mean of the background noise across
            the observational image.

        Returns
        -------
        A`utility.containers.ExtractionResults` instance or a
        Pandas Dataframe containing the results of the source
        extraction.

        Notes
        -----
        See Hopkins et al., AJ, 123, 1086 (2002) for more details.
        http://adsabs.harvard.edu/abs/2002AJ....123.1086H

        """
        # The correlation length in config.py is used not only for the
        # calculation of error bars with the Condon formulae, but also for
        # calculating the number of independent pixels.
        corlengthlong, corlengthshort = self.correlation_lengths

        C_n = (
            1.0
            / np.arange(
                round(0.25 * np.pi * corlengthlong * corlengthshort + 1)
            )[1:]
        ).sum()

        # Calculate the FDR threshold
        # Things will go terribly wrong in the line below if the interpolated
        # noise values get very close or below zero. Use INTERPOLATE_ORDER=1
        # or the roundup option.
        if isinstance(bgmap, (np.ndarray, np.ma.MaskedArray)):
            if bgmap.shape != self.backmap.shape:
                raise IndexError("Background map has wrong shape")
            else:
                self.backmap = bgmap
        if isinstance(noisemap, (np.ndarray, np.ma.MaskedArray)):
            if noisemap.shape != self.rmsmap.shape:
                raise IndexError("Noisemap has wrong shape")
            if noisemap.min() < 0:
                raise ValueError("RMS noise cannot be negative")
            else:
                self.rmsmap = noisemap

        normalized_data = self.data_bgsubbed / self.rmsmap

        n1 = np.sqrt(2 * np.pi)
        prob = np.sort(np.ravel(np.exp(-0.5 * normalized_data**2) / n1))
        lengthprob = float(len(prob))
        compare = (alpha / C_n) * np.arange(lengthprob + 1)[1:] / lengthprob
        # Find the last undercrossing, see, e.g., fig. 9 in Miller et al., AJ
        # 122, 3492 (2001).  Searchsorted is not used because the array is not
        # sorted.
        try:
            index = (np.where(prob - compare < 0.0)[0]).max()
        except ValueError:
            # Everything below threshold
            return containers.ExtractionResults()

        fdr_threshold = np.sqrt(-2.0 * np.log(n1 * prob[index]))
        # Default we require that all source pixels are above the threshold,
        # not only the peak pixel.  This gives a better guarantee that indeed
        # the fraction of false positives is less than fdr_alpha in config.py.
        # See, e.g., Hopkins et al., AJ 123, 1086 (2002).
        if not self.conf.image.analysis_thr:
            anl = fdr_threshold
        else:
            anl = self.conf.image.analysis_thr
        return self._pyse(
            fdr_threshold * self.rmsmap,
            anl * self.rmsmap,
        )

    @staticmethod
    def box_slice_about_pixel(x, y, box_radius):
        """Returns a slice centred about (x,y), of width = 2 *
        int(box_radius) + 1.

        Parameters
        ----------
        x : int
            Desired row index.
        y : int
            Desired column index.
        box_radius : float
            Radius of the box in pixel coordinates.

        Returns
        -------
        tuple of slice
            Slice centred about (x,y) with width = 2*box_radius + 1.

        """
        ibr = int(box_radius)
        x = int(x)
        y = int(y)
        return (slice(x - ibr, x + ibr + 1), slice(y - ibr, y + ibr + 1))

    def fit_to_point(
        self, x: int, y: int, boxsize: int, threshold: float, fixed: str
    ):
        """Fit an elliptical Gaussian to a specified point on the
        image.

        Parameters
        ----------
        x : int
            Pixel x-coordinate of the point to fit.
        y : int
            Pixel y-coordinate of the point to fit.
        boxsize : int
            Length of the square section of the image to use for the fit.
        threshold : float
            Threshold below which data is not used for fitting (in units of
            rmsmap).
        fixed : str
            If set to ``position``, the pixel coordinates are fixed in the
            fit.

        Returns
        -------
        Detection
            An instance of :class:`sourcefinder.extract.Detection` containing
            the fit results.

        """
        logger.debug("Force-fitting pixel location ({},{})".format(x, y))
        # First, check that x and y are actually valid semi-positive integers.
        # Otherwise,
        # If they are too high (positive), then indexing will fail
        # BUT, if they are negative, then we get wrap-around indexing
        # and the fit continues at the wrong position!
        if x < 0 or x > self.xdim or y < 0 or y > self.ydim:
            logger.warning(
                "Dropping forced fit at ({},{}), "
                "pixel position outside image".format(x, y)
            )
            return None

        # Next, check if any of the central pixels (in a 3x3 box about the
        # fitted pixel position) have been Masked
        # (e.g. if NaNs, or close to image edge) - reject if so.
        central_pixels_slice = ImageData.box_slice_about_pixel(x, y, 1)
        if self.data.mask[central_pixels_slice].any():
            logger.warning(
                "Dropping forced fit at ({},{}), "
                "Masked pixel in central fitting region".format(x, y)
            )
            return None

        if (
            # Recent NumPy
            hasattr(np.ma.core, "MaskedConstant")
            and isinstance(self.rmsmap, np.ma.core.MaskedConstant)
        ) or (
            # Old NumPy
            np.ma.is_masked(self.rmsmap[int(x), int(y)])
        ):
            logger.error("Background is masked: cannot fit")
            return None

        chunk = ImageData.box_slice_about_pixel(x, y, boxsize / 2.0)
        if threshold is not None:
            # We'll mask out anything below threshold*self.rmsmap from the fit.
            labels, num = self.labels.setdefault(
                # Dictionary mapping threshold -> islands map
                threshold,
                ndimage.label(
                    self.clip.setdefault(
                        # Dictionary mapping threshold -> mask
                        threshold,
                        np.where(
                            self.data_bgsubbed > threshold * self.rmsmap, 1, 0
                        ),
                    )
                ),
            )

            mylabel = labels[int(x), int(y)]
            if mylabel == 0:  # 'Background'
                raise ValueError(
                    "Fit region is below specified threshold, fit aborted."
                )
            mask = np.where(labels[chunk] == mylabel, 0, 1)
            fitme = np.ma.array(self.data_bgsubbed[chunk], mask=mask)
            if len(fitme.compressed()) < 1:
                raise IndexError("Fit region too close to edge or too small")
        else:
            fitme = self.data_bgsubbed[chunk]
            if fitme.size < 1:
                raise IndexError("Fit region too close to edge or too small")

        if not len(fitme.compressed()):
            logger.error("All data is masked: cannot fit")
            return None

        # set argument for fixed parameters based on input string
        if fixed == "position":
            _fixed = {"xbar": boxsize / 2.0, "ybar": boxsize / 2.0}
        elif fixed == "position+shape":
            _fixed = {
                "xbar": boxsize / 2.0,
                "ybar": boxsize / 2.0,
                "semimajor": self.beam[0],
                "semiminor": self.beam[1],
                "theta": self.beam[2],
            }
        elif fixed is None:
            _fixed = {}
        else:
            raise TypeError("Unkown fixed parameter")

        if threshold is not None:
            threshold_at_pixel = threshold * self.rmsmap[int(x), int(y)]
        else:
            threshold_at_pixel = None

        try:
            measurement, _, _ = extract.source_profile_and_errors(
                fitme,
                threshold_at_pixel,
                self.rmsmap[chunk],
                self.rmsmap[int(x), int(y)],
                self.beam,
                self.fudge_max_pix_factor,
                self.beamsize,
                self.correlation_lengths,
                self.conf.image,
                fixed=_fixed,
            )
        except ValueError:
            # Fit failed to converge
            # Moments are not applicable when holding parameters fixed
            logger.error("Gaussian fit failed at %f, %f", x, y)
            return None

        try:
            assert abs(measurement["xbar"]) < boxsize
            assert abs(measurement["ybar"]) < boxsize
        except AssertionError:
            logger.warning("Fit falls outside of box.")

        measurement["xbar"] += x - boxsize / 2.0
        measurement["ybar"] += y - boxsize / 2.0
        measurement.sig = (fitme / self.rmsmap[chunk]).max()

        return extract.Detection(measurement, self)

    def fit_fixed_positions(
        self,
        positions,
        boxsize,
        threshold=None,
        fixed="position+shape",
        ids=None,
    ):
        """Convenience function to fit a list of sources at the given
        positions.

        This function wraps around :py:func:`fit_to_point`.

        Parameters
        ----------
        positions : list of tuples
            List of (RA, Dec) tuples. Positions to be fit, in decimal degrees.
        boxsize : int
            Length of the square section of the image to use for the fit.
        threshold : float, default: None
            Threshold below which data is not used for fitting.
        fixed : str, default: 'position+shape'
            If set to `position`, the pixel coordinates are fixed in the fit.
        ids : tuple, default: None
            List of identifiers. If not None, must match the length and order of
            the requested fits.

        Note
        ----
        boxsize is in pixel coordinates, not in sky coordinates.

        Returns
        -------
        tuple
            A list of successful fits. If ``ids`` is None, returns a single list
            of :class:`sourcefinder.extract.Detection` s. Otherwise, returns a
            tuple of two matched lists: ([detections], [matching_ids]).

        """
        if ids is not None:
            assert len(ids) == len(positions)

        successful_fits = []
        successful_ids = []
        for idx, posn in enumerate(positions):
            try:
                (
                    x,
                    y,
                ) = self.wcs.s2p((posn[0], posn[1]))
            except RuntimeError as e:
                if str(e).startswith("wcsp2s error: 8:") or str(e).startswith(
                    "wcsp2s error: 9:"
                ):
                    logger.warning(
                        "Input coordinates (%.2f, %.2f) invalid: ",
                        posn[0],
                        posn[1],
                    )
                else:
                    raise
            else:
                try:
                    fit_results = self.fit_to_point(
                        x, y, boxsize=boxsize, threshold=threshold, fixed=fixed
                    )
                    if not fit_results:
                        # We were unable to get a good fit
                        continue
                    if fit_results.ra.error == float(
                        "inf"
                    ) or fit_results.dec.error == float("inf"):
                        logging.warning("position errors extend outside image")
                    else:
                        successful_fits.append(fit_results)
                        if ids:
                            successful_ids.append(ids[idx])

                except IndexError as e:
                    logger.warning(
                        "Input pixel coordinates (%.2f, %.2f) "
                        "could not be fit because: " + str(e),
                        posn[0],
                        posn[1],
                    )
        if ids:
            return successful_fits, successful_ids
        return successful_fits

    def label_islands(self, detectionthresholdmap, analysisthresholdmap):
        """Return a labelled array of pixels for fitting.

        Parameters
        ----------
        detectionthresholdmap : np.ma.MaskedArray
            Detection threshold map with shape (nrow, ncol), matching the shape
            of the observational image (self.rawdata). The values are of dtype
            np.float32.
        analysisthresholdmap : np.ma.MaskedArray
            Analysis threshold map with shape (nrow, ncol), matching the shape
            of the observational image (self.rawdata). The values are of dtype
            np.float32.

        Returns
        -------
        tuple
            - labels_above_det_thr (np.ndarray): 1D array of labels
              above detection threshold, with shape
              (num_islands_above_detection_threshold,) and dtype
              np.int64.  Note that the length of this array may be
              smaller than the total number of islands above the
              analysis threshold, as some labels may have been
              filtered out due to a peak spectral brightness lower
              than the local detection threshold.

            - labelled_data (np.ndarray): Array of labelled pixels,
              where each pixel with a nonzero label corresponds to an
              island above the analysis threshold. The array has the
              same shape as the observational image (self.rawdata) and
              contains integer values corresponding to the labels of
              the islands. Pixels that do not belong to any island are
              assigned a label of 0. The number of islands above the
              analysis threshold is equal to the number of unique
              labels in this array, which is equal to or larger than
              num_islands_above_detection_threshold, i.e. the number
              of islands above the detection threshold.  This array
              has dtype np.int32.

            - num_islands_above_detection_threshold (int): Number of
              islands above detection threshold.

            - maxposs_above_det_thr (np.ndarray): Array of indices of
              the maximum pixel values above detection threshold, with
              shape (num_islands_above_detection_threshold, 2) and
              dtype np.int32.

            - maxis_above_det_thr (np.ndarray): Array of maximum pixel
              values above detection threshold, with shape
              (num_islands_above_detection_threshold,) and dtype
              np.float32.

            - npixs_above_det (np.ndarray): 1D array of pixel counts
              for each island with peak spectral brightness above the
              detection threshold, with shape
              (num_islands_above_detection_threshold,) and dtype
              np.int32.

            - all_indices_above_det_thr (np.ndarray): Array of indices
              of the islands above detection threshold, with shape
              (num_islands_above_detection_threshold, 4) and dtype
              np.int32.

            - slices (list): List of slices encompassing all islands
              in labelled_data, i.e.  encompassing all islands above
              the analysis threshold.

        """
        # If there is no usable data, we return an empty set of islands.
        if not len(self.rmsmap.compressed()):
            logging.warning("RMS map masked; sourcefinding skipped")
            return (
                [],
                np.zeros(self.data_bgsubbed.shape, dtype=np.int32),
                None,
                None,
                None,
                None,
                None,
                None,
            )

        # At this point, we select all the data which is eligible for
        # source fitting. We are actually using three separate filters, which
        # exclude:
        #
        # 1. Anything which has been masked before we reach this point;
        # 2. Any pixels which fall below the analysis threshold at that pixel
        #    position;
        # 3. Any pixels corresponding to a position where the RMS noise is
        #    less than RMS_FILTER times the median RMS across
        #    the whole image.
        #
        # The third filter attempts to exclude those regions of the image
        # which contain no usable data; for example, the parts of the image
        # falling outside the circular region produced by awimager.

        if self.conf.image.rms_filter:
            clipped_data = np.ma.where(
                (self.data_bgsubbed > analysisthresholdmap)
                & (
                    self.rmsmap
                    >= (
                        self.conf.image.rms_filter
                        * np.median(self.grids["rms"])
                    )
                ),
                1,
                0,
            ).filled(fill_value=0)
        else:
            clipped_data = np.ma.where(
                self.data_bgsubbed > analysisthresholdmap, 1, 0
            ).filled(fill_value=0)

        labelled_data, num_labels = ndimage.label(
            clipped_data, self.conf.image.structuring_element
        )

        # Get a bounding box for each island:
        # NB Slices ordered by label value (1...N,)
        # 'None' returned for missing label indices.
        slices = ndimage.find_objects(labelled_data)

        # Derive all indices than correspond to the slices
        all_indices = ImageData.slices_to_indices(slices)

        # Derive maximum positions, maximum values and number of pixels
        # per island.
        maxposs = np.empty((num_labels, 2), dtype=np.int32)
        dummy = np.empty_like(maxposs)
        maxis = np.empty(num_labels, dtype=np.float32)
        npixs = np.empty(num_labels, dtype=np.int32)

        ImageData.extract_parms_image_slice(
            self.data_bgsubbed.data.astype(dtype=np.float32, copy=False),
            all_indices,
            labelled_data,
            np.arange(1, num_labels + 1, dtype=np.int32),
            dummy,
            maxposs,
            maxis,
            npixs,
        )

        # Here we remove the labels that correspond to islands below the
        # detection threshold.
        above_det_thr = (
            maxis > detectionthresholdmap[maxposs[:, 0], maxposs[:, 1]]
        )

        num_islands_above_detection_threshold = above_det_thr.sum()

        # np.arange(1, num_labels + 1)) to discard the zero label
        # (background).
        # Will break in the pathological case that all the image pixels
        # are covered by sources, but we will take that risk.
        labels_above_det_thr = np.extract(
            above_det_thr, np.arange(1, num_labels + 1)
        )

        maxposs_above_det_thr = np.compress(above_det_thr, maxposs, axis=0)
        maxis_above_det_thr = np.extract(above_det_thr, maxis)
        npixs_above_det = np.extract(above_det_thr, npixs)
        all_indices_above_det_thr = np.compress(
            above_det_thr, all_indices, axis=0
        )

        print(
            f"Number of detected sources ="
            f" {num_islands_above_detection_threshold}"
        )

        return (
            labels_above_det_thr,
            labelled_data,
            num_islands_above_detection_threshold,
            maxposs_above_det_thr,
            maxis_above_det_thr,
            npixs_above_det,
            all_indices_above_det_thr,
            slices,
        )

    @staticmethod
    def fit_islands(
        fudge_max_pix_factor, beamsize, correlation_lengths, fixed, island
    ):
        """This function was created to enable the use of 'partial'
        such that we can parallellize source measurements

        """
        return island.fit(
            fudge_max_pix_factor, beamsize, correlation_lengths, fixed=fixed
        )

    # “The sliced_to_indices function has been accelerated by a factor 2.7
    # using ChatGPT 4.0.  Its AI-output has been verified for correctness,
    # accuracy and completeness, adapted where needed, and approved by the
    # author
    @staticmethod
    def slices_to_indices(slices):
        """Convert the list of tuples of slices generated by
        scipy.ndimage.find_objects into a 2D int32 array with number
        of rows equal to the number of islands and 4 columns, i.e 4
        integers per island, containing the same information as the
        slices, but more suitable for compilation by Numba

        """
        num_slices = len(slices)
        all_indices = np.empty((num_slices, 4), dtype=np.int32)

        # Extract start and stop indices for rows and columns
        all_indices[:, 0] = [s[0].start for s in slices]  # Row start
        all_indices[:, 1] = [s[0].stop for s in slices]  # Row stop
        all_indices[:, 2] = [s[1].start for s in slices]  # Column start
        all_indices[:, 3] = [s[1].stop for s in slices]  # Column stop

        return all_indices

    @staticmethod
    @guvectorize(
        [
            (
                float32[:, :],
                int32[:],
                int32[:, :],
                int32[:],
                int32[:],
                int32[:],
                float32[:],
                int32[:],
            )
        ],
        "(n, m), (l), (n, m), " + "(), (k) -> (k), (), ()",
    )
    def extract_parms_image_slice(
        some_image, inds, labelled_data, label, dummy, maxpos, maxi, npix
    ):
        """Find the highest pixel value and its position.

        For an island, indicated by a group of pixels with the same label,
        find the highest pixel value and its position, first relative to the
        upper left corner of the rectangular slice encompassing the island,
        but finally relative to the upper left corner of the image, i.e. the
        [0, 0] position of the Numpy array with all the image pixel values.
        Also, derive the number of pixels of the island.

        Parameters
        ----------
        some_image : np.ndarray
            2D array with all the pixel values, typically
            self.data_bgsubbed.data.
        inds : np.ndarray
            Array of four indices indicating the slice encompassing an island.
            Such a slice would typically be a pick from a list of slices from a
            call to scipy.ndimage.find_objects. Since we are attempting
            vectorized processing here, the slice should have been replaced by
            its four coordinates through a call to slices_to_indices.
        labelled_data : np.ndarray
            Array with the same shape as some_image, with labelled islands with
            integer values and zeroes for all background pixels.
        label : int
            The label (integer value) corresponding to the slice encompassing
            the island. Or actually it should be the other way round, since
            there can be multiple islands within one rectangular slice.
        dummy : np.ndarray
            Artefact of the implementation of guvectorize: Empty array with the
            same shape as maxpos. It is needed because of a missing feature in
            guvectorize: There is no other way to tell guvectorize what the
            shape of the output array will be. Therefore, we define an
            otherwise redundant input array with the same shape as the
            desired output array. Defined as int32, but could be any type.
        maxpos : np.ndarray
            Array of two integers indicating the indices of the highest pixel
            value of the island with label = label relative to the position of
            pixel [0, 0] of the image.
        maxi : np.float32
            Float32 equal to the highest pixel value of the island with
            label=label.
        npix : np.int32
            Integer indicating the number of pixels of the island.

        Returns
        -------
        None
            No return values, because of the use of the guvectorize decorator:
            'guvectorize() functions don’t return their result value: they take
            it as an array argument, which must be filled in by the function'. In
            this case maxpos, maxi and npix will be filled with values.

        """

        labelled_data_chunk = labelled_data[
            inds[0] : inds[1], inds[2] : inds[3]
        ]
        image_chunk = some_image[inds[0] : inds[1], inds[2] : inds[3]]
        segmented_island = np.where(labelled_data_chunk == label[0], 1, 0)

        selected_data = segmented_island * image_chunk
        maxpos_flat = selected_data.argmax()
        maxpos[0] = np.floor_divide(maxpos_flat, selected_data.shape[1])
        maxpos[1] = np.mod(maxpos_flat, selected_data.shape[1])
        maxi[0] = selected_data[maxpos[0], maxpos[1]]
        maxpos[0] += inds[0]
        maxpos[1] += inds[2]
        npix[0] = int32(segmented_island.sum())

    def _pyse(
        self,
        detectionthresholdmap,
        analysisthresholdmap,
        labelled_data=None,
        labels=np.array([], dtype=np.int32),
    ):
        """Run Python-based source extraction on this image.

        Parameters
        ----------
        detectionthresholdmap : np.ma.MaskedArray
            2D array of floats with the same shape as the observational image
            (self.rawdata). The detection threshold map imposes an extra
            threshold for source detection and is therefore higher than the
            analysis threshold map.
        analysisthresholdmap : np.ma.MaskedArray
            2D array of floats with the same shape as the observational image
            (self.rawdata). analysisthresholdmap imposes the primary threshold
            for source detection. All the pixels within the island that exceed
            this will be used when measuring the source. It is lower (or equal)
            than detectionthresholdmap, or else we would be left with too few
            pixels for proper source shape measurements, in some cases. This
            map is computed as analysis_threshold * self.rmsmap.
        labelled_data : np.ndarray, optional, default=None
            Labelled island map (output of np.ndimage.label()). Will be
            calculated automatically if not provided.
        labels : np.ndarray, optional, default=np.array([], dtype=np.int32)
            Array of integers representing the labels in the island map to use
            for fitting.

        Returns
        -------
        A `utility.containers.ExtractionResults` instance or a Pandas
        DataFrame containing the results of the source extraction.

        Notes
        -----
        This is described in detail in the "LOFAR Transients Pipeline" article
        by John D. Swinbank et al., see
        https://doi.org/10.1016/j.ascom.2015.03.002

        """
        # Map our chunks onto a list of islands.

        if labelled_data is None:
            (
                labels,
                labelled_data,
                num_islands,
                maxposs,
                maxis,
                npixs,
                indices,
                slices,
            ) = self.label_islands(detectionthresholdmap, analysisthresholdmap)

        num_islands = len(labels)

        if self.conf.image.deblend_nthresh or not self.conf.image.vectorized:
            results = containers.ExtractionResults()
            island_list = []
            for label in labels:
                chunk = slices[label - 1]
                # In selected_data only the pixels with the "correct"
                # (see above) labels are retained. Other pixel values are
                # set to -(bignum).
                # In this way, disconnected pixels within (rectangular)
                # slices around islands (particularly the large ones) do
                # not affect the source measurements.
                selected_data = np.ma.where(
                    labelled_data[chunk] == label,
                    self.data_bgsubbed[chunk].data,
                    -extract.BIGNUM,
                ).filled(fill_value=-extract.BIGNUM)

                island_list.append(
                    extract.Island(
                        selected_data,
                        self.rmsmap[chunk],
                        chunk,
                        detectionthresholdmap[chunk],
                        self.beam,
                        self.conf.image,
                    )
                )

            if self.conf.export.islands:
                self.Gaussian_islands = np.zeros(self.data.shape)
            # If required, we can save the 'leftovers' from the deblending and
            # fitting processes for later analysis. This needs setting up here:
            if self.conf.export.residuals:
                self.Gaussian_residuals = np.zeros(self.data.shape)
                self.residuals_from_deblending = np.zeros(self.data.shape)
                for island in island_list:
                    self.residuals_from_deblending[
                        island.chunk
                    ] += island.data.filled(fill_value=0.0)

            # Deblend each of the islands to its consituent parts, if necessary
            if self.conf.image.deblend_nthresh:
                deblended_list = [x.deblend() for x in island_list]
                island_list = list(utils.flatten(deblended_list))

            # Set up the fixed fit parameters if 'force beam' is on:
            if self.conf.image.force_beam:
                fixed = {
                    "semimajor": self.beam[0],
                    "semiminor": self.beam[1],
                    "theta": self.beam[2],
                }
            else:
                fixed = None

            # Iterate over the list of islands and measure the source in each,
            # appending it to the results list.
            fit_islands_partial = partial(
                ImageData.fit_islands,
                self.fudge_max_pix_factor,
                self.beamsize,
                self.correlation_lengths,
                fixed,
            )

            max_workers = self.conf.image.nr_threads
            with ThreadPoolExecutor(max_workers=max_workers) as executor:
                fit_results = list(
                    executor.map(fit_islands_partial, island_list)
                )

            for island, fit_result in zip(island_list, fit_results):
                if fit_result:
                    measurement, Gauss_island, Gauss_residual = fit_result
                else:
                    # Failed to fit; drop this island and go to the next.
                    continue
                try:
                    det = extract.Detection(
                        measurement,
                        self,
                        chunk=island.chunk,
                    )
                    if det.ra.error == float("inf") or det.dec.error == float(
                        "inf"
                    ):
                        logger.warning(
                            (
                                "Bad fit from blind extraction at "
                                "pixel coords:%f %f - measurement "
                                "discarded "
                                "(increase fitting margin?)"
                            ),
                            det.x,
                            det.y,
                        )
                    else:
                        results.append(det)
                except RuntimeError as e:
                    logger.error("Island not processed; unphysical?")

                if self.conf.export.islands:
                    self.Gaussian_islands[island.chunk] += Gauss_island

                if self.conf.export.residuals:
                    self.residuals_from_deblending[
                        island.chunk
                    ] -= island.data.filled(fill_value=0.0)
                    self.Gaussian_residuals[island.chunk] += Gauss_residual

        else:

            (
                moments_of_sources,
                sky_barycenters,
                ra_errors,
                dec_errors,
                error_radii,
                smaj_asec,
                errsmaj_asec,
                smin_asec,
                errsmin_asec,
                theta_celes_values,
                theta_celes_errors,
                theta_dc_celes_values,
                theta_dc_celes_errors,
                Gaussian_islands,
                Gaussian_residuals,
                sig,
                chisq,
                reduced_chisq,
            ) = extract.source_measurements_vectorised(
                num_islands,
                npixs,
                maxposs,
                maxis,
                self.data_bgsubbed.data,
                self.rmsmap.data,
                analysisthresholdmap.data,
                indices,
                labelled_data,
                labels,
                self.wcs,
                self.fudge_max_pix_factor,
                self.beam,
                self.beamsize,
                self.correlation_lengths,
                self.conf,
            )

            if self.conf.export.islands:
                self.Gaussian_islands = Gaussian_islands

            if self.conf.export.residuals:
                self.Gaussian_residuals = Gaussian_residuals

            if not self.conf.export.reconvert:
                sources_df = make_measurements_dataframe(
                    moments_of_sources,
                    sky_barycenters,
                    ra_errors,
                    dec_errors,
                    error_radii,
                    smaj_asec,
                    errsmaj_asec,
                    smin_asec,
                    errsmin_asec,
                    theta_celes_values,
                    theta_celes_errors,
                    theta_dc_celes_values,
                    theta_dc_celes_errors,
                    sig,
                    chisq,
                    reduced_chisq,
                )
                return sources_df[self.conf.export.source_params]
            else:
                results = containers.ExtractionResults()
                for count, label in enumerate(labels):
                    chunk = slices[label - 1]

                    param = extract.ParamSet()
                    param.sig = sig[count]
                    param.chisq = chisq[count]
                    param.reduced_chisq = reduced_chisq[count]

                    param["peak"] = Uncertain(
                        moments_of_sources[count, 0, 0],
                        moments_of_sources[count, 1, 0],
                    )
                    param["flux"] = Uncertain(
                        moments_of_sources[count, 0, 1],
                        moments_of_sources[count, 1, 1],
                    )
                    param["xbar"] = Uncertain(
                        moments_of_sources[count, 0, 2],
                        moments_of_sources[count, 1, 2],
                    )
                    param["ybar"] = Uncertain(
                        moments_of_sources[count, 0, 3],
                        moments_of_sources[count, 1, 3],
                    )
                    param["semimajor"] = Uncertain(
                        moments_of_sources[count, 0, 4],
                        moments_of_sources[count, 1, 4],
                    )
                    param["semiminor"] = Uncertain(
                        moments_of_sources[count, 0, 5],
                        moments_of_sources[count, 1, 5],
                    )
                    param["theta"] = Uncertain(
                        moments_of_sources[count, 0, 6],
                        moments_of_sources[count, 1, 6],
                    )
                    param["semimaj_deconv"] = Uncertain(
                        moments_of_sources[count, 0, 7],
                        moments_of_sources[count, 1, 7],
                    )
                    param["semimin_deconv"] = Uncertain(
                        moments_of_sources[count, 0, 8],
                        moments_of_sources[count, 1, 8],
                    )
                    param["theta_deconv"] = Uncertain(
                        moments_of_sources[count, 0, 9],
                        moments_of_sources[count, 1, 9],
                    )

                    det = extract.Detection(param, self, chunk=chunk)
                    results.append(det)

        def is_usable(det):
            """Check that both ends of each axis are usable.

            I.e., they fall within an unmasked part of the image. The
            axis will not likely fall exactly on a pixel number, so
            check all the surroundings.

            """

            def check_point(x, y):
                x = (int(x), int(np.ceil(x)))
                y = (int(y), int(np.ceil(y)))
                for position in itertools.product(x, y):
                    try:
                        if self.data.mask[position[0], position[1]]:
                            # Point falls in mask
                            return False
                    except IndexError:
                        # Point falls completely outside image
                        return False
                # Point is ok
                return True

            for point in (
                (det.start_smaj_x, det.start_smaj_y),
                (det.start_smin_x, det.start_smin_y),
                (det.end_smaj_x, det.end_smaj_y),
                (det.end_smin_x, det.end_smin_y),
            ):
                if not check_point(*point):
                    logger.debug(
                        "Unphysical source at pixel %f, %f"
                        % (det.x.value, det.y.value)
                    )
                    return False
            return True

        # Filter will return a list; ensure we return an ExtractionResults.
        return containers.ExtractionResults(list(filter(is_usable, results)))



import logging
import warnings
from math import degrees

from casacore.tables import table as casacore_table

from sourcefinder.utils import is_valid_beam_tuple
from sourcefinder.accessors.dataaccessor import DataAccessor
from sourcefinder.utility.coordinates import WCS
from sourcefinder.utility.coordinates import mjd2datetime

logger = logging.getLogger(__name__)


class CasaImage(DataAccessor):
    """
    Provides common functionality for pulling data from the CASA image format.

    (Technically known as the 'MeasurementSet' format.)

    Parameters
    ----------
    url : str
        The location of the CASA image file.
    plane : int, default: 0
        The data plane to extract from the image cube.
    beam : tuple, default: None
        Beam parameters (major axis, minor axis, position angle) in degrees.
        If not provided, these are extracted from the CASA table.

    Notes
    -----
    CasaImage does not provide tau_time or taustart_ts, as there was
    no clear standard for these metadata, so cannot be
    instantiated directly - subclass it and extract these attributes
    as appropriate to a given telescope.
    """

    def __init__(self, url, plane=0, beam=None):
        # super().__init__()
        self.url = url

        # we don't want the table as a property since it makes the accessor
        # not serializable
        table = casacore_table(self.url, ack=False)
        self.data = self.parse_data(table, plane)
        self.wcs = self.parse_coordinates(table)
        self.centre_ra, self.centre_decl = self.parse_phase_centre(table)
        self.freq_eff, self.freq_bw = self.parse_frequency(table)
        self.pixelsize = self.parse_pixelsize()

        if is_valid_beam_tuple(beam) or not is_valid_beam_tuple(self.beam):
            # An argument-supplied beam overrides a beam derived from
            # (bmaj, bmin, bpa) in a config.toml. Only if those two options
            # are not specified, we parse the beam from the header.
            bmaj, bmin, bpa = beam if is_valid_beam_tuple(beam) else (
                self.parse_beam(table))
            self.beam = self.degrees2pixels(
                bmaj, bmin, bpa, self.pixelsize[0], self.pixelsize[1]
            )

    def parse_data(self, table, plane=0):
        """
        Extract and squeeze data from a CASA table, select desired plane and
        transpose.
        Squeezing is done to remove any singleton dimensions.

        Parameters
        ----------
        table : casacore.tables.table
            The CASA table from which data is extracted.
        plane : int, default: 0
            The data plane to extract from the image cube.

        Returns
        -------
        data : numpy.ndarray
            The extracted and transposed data array. If the data cube has more
            than two dimensions - after removing singleton dimensions - the
            plane specified by the `plane` parameter is selected from the first
            axis.
        """
        data = table[0]['map'].squeeze()
        planes = len(data.shape)
        if planes != 2:
            msg = ("received datacube with %s planes, assuming Stokes I "
                   "and taking plane 0" % planes)
            logger.warning(msg)
            warnings.warn(msg)
            data = data[plane, :, :]
        data = data.transpose()
        return data

    def parse_coordinates(self, table):
        """
        Parse and return a WCS (World Coordinate System) object from the
        CASA table.

        Parameters
        ----------
        table : casacore.tables.table
            The CASA table from which the WCS information is extracted.

        Returns
        -------
        wcs : WCS
            A WCS object containing the coordinate system information extracted
            from the CASA table.
        """
        wcs = WCS()
        my_coordinates = table.getkeyword('coords')['direction0']
        wcs.crval = my_coordinates['crval']
        wcs.crpix = my_coordinates['crpix']
        wcs.cdelt = my_coordinates['cdelt']
        ctype = ['unknown', 'unknown']
        # What about other projections?!
        if my_coordinates['projection'] == "SIN":
            if my_coordinates['axes'][0] == "Right Ascension":
                ctype[0] = "RA---SIN"
            if my_coordinates['axes'][1] == "Declination":
                ctype[1] = "DEC--SIN"
        wcs.ctype = tuple(ctype)
        # Rotation, units? We better set a default
        wcs.crota = (0., 0.)
        wcs.cunit = table.getkeyword('coords')['direction0']['units']
        return wcs

    def parse_frequency(self, table):
        """
        Extract frequency-related information from the CASA table headers.

        Parameters
        ----------
        table : casacore.tables.table
            The CASA table from which frequency information is extracted.

        Returns
        -------
        freq_eff : float
            The effective frequency (rest frequency) in Hz.
        freq_bw : float
            The frequency bandwidth (channel width) in Hz.
        """
        freq_eff = table.getkeywords()['coords']['spectral2']['restfreq']
        freq_bw = table.getkeywords()['coords']['spectral2']['wcs']['cdelt']
        return freq_eff, freq_bw

    def parse_beam(self, table):
        """Extract beam parameters from the CASA table.

        Parameters
        ----------
        table : casacore.tables.table
            The CASA table from which the beam parameters are extracted.

        Returns
        -------
        bmaj : float
            The semi-major axis of the beam in degrees.
        bmin : float
            The semi-minor axis of the beam in degrees.
        bpa : float
            The angle of the major axis of the synthesized beam,
            measured east from local north, in degrees.

        """

        def ensure_degrees(quantity):
            """
            Convert a quantity to degrees.

            Parameters
            ----------
            quantity : dict
                A dictionary containing the value and unit of the quantity to
                be converted. Supported units are 'deg', 'arcsec', and 'rad'.

            Returns
            -------
            float
                The value of the quantity converted to degrees.

            Raises
            ------
            Exception
                If the unit of the quantity is not recognized.
            """
            if quantity['unit'] == 'deg':
                return quantity['value']
            elif quantity['unit'] == 'arcsec':
                return quantity['value'] / 3600
            elif quantity['unit'] == 'rad':
                return degrees(quantity['value'])
            else:
                raise Exception("Beam units (%s) unknown" % quantity['unit'])

        restoringbeam = table.getkeyword('imageinfo')['restoringbeam']
        bmaj = ensure_degrees(restoringbeam['major'])
        bmin = ensure_degrees(restoringbeam['minor'])
        bpa = ensure_degrees(restoringbeam['positionangle'])
        return bmaj, bmin, bpa

    @staticmethod
    def parse_phase_centre(table):
        """
        Assume the units for the pointing centre are in radians and return the
        RA modulo 360.

        Parameters
        ----------
        table : casacore.tables.table
            The CASA table from which the pointing centre is extracted.

        Notes
        -----
        The units for the pointing centre are not given in either the image
        cube itself nor in the image content description. Assume radians. The
        Right Ascension (RA) is returned modulo 360 to ensure it is always
        within the range 0 <= RA < 360.
        """
        centre_ra, centre_decl = table.getkeyword('coords')['pointingcenter'][
            'value']
        return degrees(centre_ra) % 360, degrees(centre_decl)

    @staticmethod
    def parse_taustartts(table):
        """
        Extract integration start-time from CASA table header.

        This applies to some CASA images (typically those created from uvFITS
        files) but not all, and so should be called for each sub-class.

        Parameters
        ----------
        table : casacore.tables.table
            The MAIN table of the CASA image.

        Returns
        -------
        datetime.datetime
            Time of observation start as an instance of `datetime.datetime`.
        """
        obsdate = table.getkeyword('coords')['obsdate']['m0']['value']
        return mjd2datetime(obsdate)

    @staticmethod
    def unique_column_values(table, column_name):
        """
        Find all the unique values in a particular column of a CASA table.

        Parameters
        ----------
        table : casacore.tables.table
            The CASA table from which the column values are extracted.
        column_name : str
            The name of the column to extract unique values from.

        Returns
        -------
        numpy.ndarray
            An array containing the unique values in the specified column.
        """
        return table.query(
            columns=column_name, sortlist="unique %s" % (column_name)
        ).getcol(column_name)



import logging

from casacore.tables import table as casacore_table

from sourcefinder.accessors.casaimage import CasaImage

logger = logging.getLogger(__name__)


class AartfaacCasaImage(CasaImage):
    """A class to represent an AARTFAAC CASA image, extending the CasaImage class.

    Parameters
    ----------
    url : str
        The location of the CASA table.
    plane : int, default: 0
        If the data is a cube, specifies which plane to use.
    beam : tuple, default: None
        Beam parameters in degrees, in the form (bmaj, bmin, bpa).

    """
    def __init__(self, url, plane=0, beam=None):
        super().__init__(url, plane=0, beam=None)
        table = casacore_table(self.url, ack=False)
        self.taustart_ts = self.parse_taustartts(table)
        self.telescope = table.getkeyword('coords')['telescope']

        # TODO: header does't contain integration time
        # aartfaac imaginig pipeline issue #25
        self.tau_time = 1

    def parse_frequency(self, table):
        """Extract frequency-related information from the casacore
        table.

        This method overrides the implementation in the `CasaImage` class,
        which retrieves the entries from the 'spectral2' sub-table.

        Parameters
        ----------
        table : casacore.tables.table
            The CASA table from which frequency information is extracted.

        Returns
        -------
        freq_eff : float
            The effective frequency (rest frequency) in Hz extracted
            from the casacore table.
        freq_bw : float
            The frequency bandwidth in Hz, derived from the WCS
            'cdelt' value in the casacore table.

        """
        keywords = table.getkeywords()

        # due to some undocumented casacore feature, the 'spectral' keyword
        # changes from spectral1 to spectral2 when AARTFAAC imaging developers
        # changed some of the header information. For now we will try both
        # locations.
        if 'spectral1' in keywords['coords']:
            keyword = 'spectral1'

        if 'spectral2' in keywords['coords']:
            keyword = 'spectral2'

        freq_eff = keywords['coords'][keyword]['restfreq']
        freq_bw = keywords['coords'][keyword]['wcs']['cdelt']
        return freq_eff, freq_bw



"""
Mock / synthetic data objects for use in testing.
"""
import numpy as np
from sourcefinder.accessors.dataaccessor import DataAccessor
from sourcefinder.utility.coordinates import WCS
import datetime

class Mock(object):
    def __init__(self, returnvalue=None):
        self.callcount = 0
        self.callvalues = []
        self.returnvalue = returnvalue

    def __call__(self, *args, **kwargs):
        self.callcount += 1
        self.callvalues.append((args, kwargs))
        return self.returnvalue


def make_wcs(crval=None,
             cdelt=None,
             crpix=None
             ):
    """
    Make a WCS object for insertion into a synthetic image.
    
    Args:
        crval : tuple, default : None
            Tuple of (RA, Dec) in decimal degrees at the reference position.
        crpix : tuple, default : None
            Tuple of (x, y) coordinates describing the reference pixel location
            corresponding to the crval sky-position.
        cdelt : tuple, default : None
            Tuple of (cdelt0, cdelt1) in decimal degrees. This is the pixel 
            width in degrees of arc, but not necessarily aligned to RA, 
            Dec unless `crota` is (0, 0). If that *is* the case, 
            then typically cdelt0 is negative since the x-axis is in the 
            direction of West (decreasing RA).
    """
    # For any arguments not set we simply assign an arbitrary valid value:
    if crval is None:
        crval = 100., 45.
    if cdelt is None:
        pixel_width_arcsec = 40
        pixel_width_deg = pixel_width_arcsec / 3600.
        cdelt = (-pixel_width_deg, pixel_width_deg)
    if crpix is None:
        crpix = (256.0, 256.0)
    wcs = WCS()
    wcs.cdelt = cdelt
    wcs.crota = (0.0, 0.0)
    wcs.crpix = crpix
    wcs.crval = crval
    wcs.ctype = ('RA---SIN', 'DEC--SIN')
    wcs.cunit = ('deg', 'deg')
    return wcs


class SyntheticImage(DataAccessor):
    def __init__(self,
                 wcs=None,
                 data=None,
                 beam=(1.5,1.5,0),
                 freq_eff=150e6,
                 freq_bw=2e6,
                 tau_time=1800,
                 taustart_ts=datetime.datetime(2015,1,1)
                 ):
        """Generate a synthetic image for use in tests.

        Parameters
        ----------
        wcs : sourcefinder.utility.coordinates.WCS, default: None
            WCS instance for the image.
        data : array_like, default: None
            Data for the image. Default is a 512x512 array of zeroes.
        beam : tuple, default: (1.5, 1.5, 0)
            Beam semi-major axis (in pixels), semi-minor axis (in pixels),
            and position angle (in radians).
        freq_eff : float, default: 150e6
            Effective frequency of the image in Hz. This is the mean frequency
            of all the visibility data which comprises this image.
        freq_bw : float, default: 2e6
            The frequency bandwidth of this image in Hz.
        tau_time : float, default: 1800
            Total time on sky in seconds.
        taustart_ts : datetime.datetime, default: datetime.datetime(2015, 1, 1)
            Timestamp of the first integration which constitutes part of this
            image.

        """
        self.url = "SyntheticImage"
        self.wcs = wcs
        if self.wcs is None:
            self.wcs = make_wcs()
        self.data = data
        if self.data is None:
            self.data = np.zeros((512,512))
        self.beam = beam
        self.freq_eff = freq_eff
        self.freq_bw = freq_bw
        self.tau_time = tau_time
        self.taustart_ts = taustart_ts

        self.pixelsize = self.parse_pixelsize()
        self.centre_ra, self.centre_decl = self.calculate_phase_centre()

    def calculate_phase_centre(self):
        """Calculate the equatorial coordinates of the center of the
        synthetic image, based on the image dimensions.

        Returns
        -------
        tuple
            A tuple containing the right ascension and declination of the
            image center in degrees.

        """
        x, y = self.data.shape
        centre_ra, centre_decl = self.wcs.p2s((x / 2, y / 2))
        return centre_ra, centre_decl



"""
This module contain utilities for the source finding routines
"""

import math
import numpy as np
from numbers import Real
import scipy.integrate
from scipy.ndimage import distance_transform_edt

from sourcefinder.gaussian import gaussian
from sourcefinder.utility import coordinates

from numba import njit, prange, guvectorize


def generate_subthresholds(min_value, max_value, num_thresholds):
    r"""Generate a series of ``num_thresholds`` logarithmically spaced values
    in the range (min_value, max_value) (both exclusive).

    First, we calculate a logarithmically spaced sequence between exp(0.0)
    and (max_value - min_value + 1). That is, the total range is between 1 and
    one greater than the difference between max_value and min_value.
    We subtract 1 from this to get the range between 0 and
    (max_value - min_value).
    We add min to that to get the range between min and max.

    This formula sets the subthreshold levels:

    .. math::

        t_i = \exp\left(
            \frac{i \cdot \log(\text{max\_value} + 1 - \text{min\_value})}
                 {\text{num\_thresholds} + 1}
        \right) + \text{min\_value} - 1

    for :math:`i = 1, 2, \ldots, \text{num\_thresholds}`.

    Parameters
    ----------
    min_value : float
        The minimum value of the range (not included in the output).
    max_value : float
        The maximum value of the range (not included in the output).
    num_thresholds : int
        The number of threshold values to generate.

    Returns
    -------
    np.ndarray
        An array of logarithmically spaced values between min_value and
        max_value (exclusive).

    """
    subthrrange = np.logspace(
        0.0,
        np.log(max_value + 1 - min_value),
        num=num_thresholds + 1,  # first value == min_value
        base=np.e,
        endpoint=False,  # do not include max_value
    )[1:]
    subthrrange += min_value - 1
    return subthrrange


def get_error_radius(wcs, x_value, x_error, y_value, y_error):
    """Estimate an absolute angular error on the position (x_value,
    y_value) with the given errors.

    Parameters
    ----------
    wcs : object
        The WCS (World Coordinate System) object used for transforming pixel
        coordinates to sky coordinates.
    x_value : float
        Position along first pixel coordinate (row index of ndarray with image
        data).
    x_error : float
        The 1-sigma error in x-value.
    y_value : float
        Position along second pixel coordinate (column index of ndarray with
        image data).
    y_error : float
        The 1-sigma error in y-value.

    Returns
    -------
    float
        The estimated absolute angular error in arcseconds.

    Notes
    -----
    This is a pessimistic estimate, because we take sum of the error
    along the X and Y axes. Better might be to project them both back on
    to the major/minor axes of the elliptical fit, but this should do for
    now.

    """
    error_radius = 0.0
    try:
        centre_ra, centre_dec = wcs.p2s([x_value, y_value])
        # We check all possible combinations in case we have a nonlinear
        # WCS.
        for pixpos in [
            (x_value + x_error, y_value + y_error),
            (x_value - x_error, y_value + y_error),
            (x_value + x_error, y_value - y_error),
            (x_value - x_error, y_value - y_error),
        ]:
            error_ra, error_dec = wcs.p2s(pixpos)
            error_radius = max(
                error_radius,
                coordinates.angsep(centre_ra, centre_dec, error_ra, error_dec),
            )
    except RuntimeError:
        # We get a runtime error from wcs.p2s if the errors place the
        # limits outside the image, in which case we set the angular
        # uncertainty to infinity.
        error_radius = float("inf")
    return error_radius


def circular_mask(xdim, ydim, radius):
    """Returns a numpy array of shape (xdim, ydim). All points within
    radius from the centre are set to 0; outside that region, they are
    set to 1.

    Parameters
    ----------
    xdim : int
        The dimension of the array along the x-axis.
    ydim : int
        The dimension of the array along the y-axis.
    radius : float
        The radius from the center within which points are set to 0.

    Returns
    -------
    np.ndarray
        A 2D ndarray with points within the radius set to 0 and outside set to
        1.

    """
    centre_x, centre_y = (xdim - 1) / 2.0, (ydim - 1) / 2.0
    x, y = np.ogrid[-centre_x : xdim - centre_x, -centre_y : ydim - centre_y]
    return x * x + y * y >= radius * radius


def generate_result_maps(data, sourcelist):
    """Return an image with Gaussian reconstructions of the sources and the
    corresponding residual image.

    Given a data array (image) and list of sources, return two images, one
    showing the sources themselves and the other the residual after the
    sources have been removed from the input data.

    Parameters
    ----------
    data : np.ndarray
        The input data array (image).
    sourcelist : list
        A list of sources to be removed from the input data.

    Returns
    -------
    gaussian_map : np.ndarray (2D)
        Shows the Gaussian reconstructions of the sources.
    residual_map : np.ndarray (2D)
        Shows the residuals from the subtractions of these
        reconstructions from the image data.

    """
    residual_map = np.array(data)  # array constructor copies by default
    gaussian_map = np.zeros(residual_map.shape)
    for src in sourcelist:
        # Include everything with 6 times the std deviation along the major
        # axis. Should be very very close to 100% of the flux.
        box_size = 6 * src.smaj.value / math.sqrt(2 * math.log(2))

        lower_bound_x = max(0, int(src.x.value - 1 - box_size))
        upper_bound_x = min(
            residual_map.shape[0], int(src.x.value - 1 + box_size)
        )
        lower_bound_y = max(0, int(src.y.value - 1 - box_size))
        upper_bound_y = min(
            residual_map.shape[1], int(src.y.value - 1 + box_size)
        )

        local_gaussian = gaussian(
            src.peak.value,
            src.x.value,
            src.y.value,
            src.smaj.value,
            src.smin.value,
            src.theta.value,
        )(
            np.indices(residual_map.shape)[
                0, lower_bound_x:upper_bound_x, lower_bound_y:upper_bound_y
            ],
            np.indices(residual_map.shape)[
                1, lower_bound_x:upper_bound_x, lower_bound_y:upper_bound_y
            ],
        )

        gaussian_map[
            lower_bound_x:upper_bound_x, lower_bound_y:upper_bound_y
        ] += local_gaussian
        residual_map[
            lower_bound_x:upper_bound_x, lower_bound_y:upper_bound_y
        ] -= local_gaussian

    return gaussian_map, residual_map


def is_valid_beam_tuple(b) -> bool:
    return (
        isinstance(b, tuple)
        and len(b) == 3
        and all(isinstance(x, Real) and x is not None for x in b)
    )


def calculate_correlation_lengths(semimajor, semiminor):
    """Calculate the Condon correlation lengths.

    In order to derive the error bars for Gaussian fits from the
    Condon (1997, PASP 109, 116C) formulae, one needs a quantity called the
    correlation length. The Condon formulae assume a circular area
    with diameter theta_N (in pixels) for the correlation: i.e. all noise
    within that area is assumed completely correlated while outside that area
    all noise is assumed completely uncorrelated. This was later generalized
    by Hopkins et al. (2003, AJ 125, 465, paragraph 3) for correlation areas
    which are not circular, i.e. for anisotropic restoring beams.

    Basically one has theta_N^2 = theta_B*theta_b.

    Good estimates in general are:

    + theta_B = 2.0 * semimajor

    + theta_b = 2.0 * semiminor

    but we have included this function to provide a convenient way of altering
    these dependencies.

    Parameters
    ----------
    semimajor : float
        The semi-major axis length in pixels.
    semiminor : float
        The semi-minor axis length in pixels.

    Returns
    -------
    tuple[float,float]
        A tuple containing the correlation lengths (theta_B, theta_b), in
        pixels.

    """

    return 2.0 * semimajor, 2.0 * semiminor


def calculate_beamsize(semimajor, semiminor):
    """Calculate the beamsize based on the semi-major and minor axes.

    Parameters
    ----------
    semimajor : float
        The semi-major axis length in pixels.
    semiminor : float
        The semi-minor axis length in pixels.

    Returns
    -------
    float
        The calculated beamsize.

    """
    return np.pi * semimajor * semiminor


def fudge_max_pix(semimajor, semiminor, theta):
    """Estimate peak spectral brightness correction at pixel of maximum spectral
    brightness.

    Previously, we adopted Rengelink's correction for the
    underestimate of the peak of the Gaussian by the maximum pixel
    method: fudge_max_pix = 1.06. See the WENSS paper
    (1997A&AS..124..259R) or his thesis.  The peak of the Gaussian
    is, of course, never at the exact center of the pixel, that's why
    the maximum pixel method will underestimate it, when averaged over an
    ensemble. This effect is smaller when the clean beam is more densely
    sampled.

    But, instead of just taking 1.06 one can make an estimate of the
    overall correction by assuming that the true peak is at a random
    position on the peak pixel and averaging over all possible
    corrections.  This overall correction makes use of the beamshape,
    so strictly speaking only accurate for unresolved sources.
    After some investigation, it turns out that this method requires not only
    unresolved sources, but also a circular beam shape. With the general
    elliptical beam, the peak pixel may be located more than half a
    pixel away from the true peak. In that case the integral below will not
    suffice as a correction. Calculating an overall correction for an ensemble
    of sources will, in the case of an elliptical beam shape, become much
    more involved.

    Parameters
    ----------
    semimajor : float
        The semi-major axis length in pixels.
    semiminor : float
        The semi-minor axis length in pixels.
    theta : float
        The position angle of the major axis in radians.

    Returns
    -------
    correction : float
        The estimated peak spectral brightness correction.

    """

    log20 = np.log(2.0)
    cos_theta = np.cos(theta)
    sin_theta = np.sin(theta)

    def landscape(y, x):
        up = math.pow(((cos_theta * x + sin_theta * y) / semiminor), 2)
        down = math.pow(((cos_theta * y - sin_theta * x) / semimajor), 2)
        return np.exp(log20 * (up + down))

    (correction, abserr) = scipy.integrate.dblquad(
        landscape, -0.5, 0.5, lambda ymin: -0.5, lambda ymax: 0.5
    )

    return correction


def flatten(nested_list):
    """Flatten a nested list

    Nested lists are made in the deblending algorithm. They're
    awful. This is a piece of code I grabbed from
    http://www.daniweb.com/code/snippet216879.html.

    The output from this method is a generator, so make sure to turn
    it into a list, like this::

        flattened = list(flatten(nested)).

    Parameters
    ----------
    nested_list : list
        A list that may contain other lists or tuples.

    Yields
    ------
    element
        The next element in the flattened list.

    Notes
    -----
        The keyword "yield" is used; i.e. a generator object is returned.

    """

    for elem in nested_list:
        if isinstance(elem, (tuple, list, np.ndarray)):
            for i in flatten(elem):
                yield i
        else:
            yield elem


# “The nearest_nonzero function has been generated using ChatGPT 4.0.
# Its AI-output has been verified for correctness, accuracy and
# completeness, adapted where needed, and approved by the author.”
def nearest_nonzero(some_arr, rms):
    """Replace values in some_arr based on the nearest non-zero
    values in rms.

    Parameters
    ----------
    some_arr : np.ndarray
        A 2D array whose values will be replaced where rms == 0.
    rms : np.ndarray
        A 2D array of the same shape as some_arr. Nearest non-zero neighbors
        in rms determine the replacement indices for some_arr.

    Returns
    -------
    np.ndarray
        A copy of some_arr with values replaced based on nearest non-zero
        neighbors in rms.

    """
    if some_arr.shape != rms.shape:
        raise ValueError("some_arr and rms must have the same shape.")

    # Handle empty array.
    if some_arr.size == 0:
        return some_arr

    # Handle single-element array.
    if some_arr.size == 1:
        if rms[0, 0] == 0:
            # Return an empty array if the single value in rms is 0.
            # A rms of 0, means a standard deviation of zero, which means we
            # cannot do any form of thresholding for source detection. If we
            # return an empty background grid, ImageData._interpolate will
            # return a completely masked background map, such that no sources
            # will be detected.
            return np.empty((0, 0))
        else:
            return some_arr  # No replacement needed if rms[0, 0] is non-zero.

    # Create a mask for zero values in rms
    zero_mask = rms == 0

    # Calculate the distance transform and nearest non-zero indices
    distances, nearest_indices = distance_transform_edt(
        zero_mask, return_indices=True
    )

    nearest_values = some_arr[nearest_indices[0], nearest_indices[1]]
    # Use nearest indices from rms to update some_arr
    result = some_arr.copy()
    result[zero_mask] = nearest_values[zero_mask]

    return result


# “The make_subimages function has been generated using ChatGPT 4.0.
# Its AI-output has been verified for correctness, accuracy and
# completeness, adapted where needed, and approved by the author.”
@njit(parallel=True)
def make_subimages(a_data, a_mask, back_size_x, back_size_y):
    """Make subimages.

    Reshape the image data such that it is suitable for guvectorized
    kappa * sigma clipping. The idea is that we have designed a function
    that will perform kappa * sigma clipping on a single flattened subimage,
    i.e. on a 1D ndarray. If we decorate that function with Numba's guvectorize
    decorator, we can use a 2D grid of subimages as input instead of a single
    flattened subimage. This means that the input to that guvectorized
    kappa * sigma clipper should be 3D. This function makes that 3D input,
    which is essentially a reshape using Numba with parallelization.

    Parameters
    ----------
    a_data: np.ndarray
        The data of the masked array (without the mask).
    a_mask: np.ndarray
        The mask of the masked array (True means the value is masked).
    back_size_x: int
        The size of the subimage along the row indices.
    back_size_y: int
        The size of the subimages along the column indices.

    Returns
    -------
    b: np.ndarray
        3D array where each subimage of size (back_size_x *
        back_size_y) is flattened and padded with NaN.
    c: np.ndarray
        2D array where each element indicates the number of unmasked
        values in the corresponding subimage.

    """
    subimage_size = back_size_x * back_size_y
    k, l = a_data.shape
    p = k // back_size_x
    r = l // back_size_y

    # Initialize b with NaNs and c with zeros
    b = np.full((p, r, subimage_size), np.nan, dtype=np.float32)
    c = np.zeros((p, r), dtype=np.int32)

    # Iterate over the subimages in parallel
    for i in prange(p):
        for j in range(r):
            # Extract the subimage (data and mask)
            subimage_data = a_data[
                i * back_size_x : (i + 1) * back_size_x,
                j * back_size_y : (j + 1) * back_size_y,
            ]
            subimage_mask = a_mask[
                i * back_size_x : (i + 1) * back_size_x,
                j * back_size_y : (j + 1) * back_size_y,
            ]

            # Preallocate an array for unmasked values (max size d*d)
            unmasked_values = np.empty(subimage_size, dtype=a_data.dtype)
            count = 0

            # Collect unmasked values manually
            for m in range(back_size_x):
                for n in range(back_size_y):
                    if not subimage_mask[m, n]:  # If not masked
                        unmasked_values[count] = subimage_data[m, n]
                        count += 1

            # Store the count of unmasked values in c
            c[i, j] = count

            # Pad the unmasked values with NaNs and store in b
            b[i, j, :count] = unmasked_values[:count]

    return b, c


# “The interp_per_row function has been generated using ChatGPT 4.0.
# Its AI-output has been verified for correctness, accuracy and
# completeness, adapted where needed, and approved by the author.”
# Define the row-wise interpolation function with guvectorize.
@guvectorize(
    ["void(float32[:], float32[:], float32[:], float32[:])"],
    "(n),(n),(k)->(k)",
    target="parallel",
    nopython=True,
    cache=True,
)
def interp_per_row(grid_row, y_initial, y_sought, interp_row):
    """Interpolate one row of the grid along the second dimension
    (y-axis).

    Parameters
    ----------
    grid_row
        1D array representing a single row of the grid.
    y_initial
        Original grid coordinates along the y-axis.
    y_sought
        Target coordinates for interpolation along the y-axis.
    interp_row
        Output array to store the interpolated row.

    """
    interp_row[:] = np.interp(y_sought, y_initial, grid_row)


# “The two_step_interp function has been generated using ChatGPT 4.0.
# Its AI-output has been verified for correctness, accuracy and
# completeness, adapted where needed, and approved by the author.”
def two_step_interp(grid, new_xdim, new_ydim):
    """Perform two-step interpolation.

    It is done on a grid to upsample it to new dimensions.  This
    function proivdes fast pieceswise bilinear interpolation in two
    steps:

    1. Interpolation across columns, i.e. per row of the input
       grid. Each row of the input grid is handled independently.
    2. Transpose the result.
    3. Again, interpolate across columns.

    This method was inspired by a comment from "tiago" in this SO
    discussion: https://stackoverflow.com/q/14530556

    Parameters
    ----------
    grid : numpy.ndarray
        The input 2D array to be upsampled.
    new_xdim : int
        The desired number of rows in the upsampled grid.
    new_ydim : int
        The desired number of columns in the upsampled grid.

    Returns
    -------
    numpy.ndarray
        The upsampled grid with dimensions (new_xdim, new_ydim).

    """
    # Define the main function for upsampling
    # Original grid coordinates
    x_initial = np.linspace(
        0, grid.shape[0] - 1, grid.shape[0], dtype=np.float32
    )
    y_initial = np.linspace(
        0, grid.shape[1] - 1, grid.shape[1], dtype=np.float32
    )

    # Target grid coordinates
    x_sought = np.linspace(
        -0.5, grid.shape[0] - 0.5, new_xdim, dtype=np.float32
    )
    y_sought = np.linspace(
        -0.5, grid.shape[1] - 0.5, new_ydim, dtype=np.float32
    )

    # Step 1: Interpolation per row.
    interp_rows = np.empty((grid.shape[0], new_ydim), dtype=np.float32)
    interp_per_row(grid, y_initial, y_sought, interp_rows)

    # Step 2: Interpolation along columns (reuse interpolate_rows)
    interp_cols = np.empty((new_xdim, new_ydim), dtype=np.float32)
    interp_per_row(interp_rows.T, x_initial, x_sought, interp_cols.T)

    return interp_cols


# “This 'newton_raphson_root_finder' function has been generated using
# ChatGPT 4.0. Its AI-output has been verified for correctness, accuracy and
# completeness, adapted where needed, and approved by the author.”
@njit
def newton_raphson_root_finder(
    f, sigma0, min_sigma, max_sigma, tol=1e-8, max_iter=100, *args
):
    """Solve the transcendental equation for sigma using Newton's
    method with interval safeguards.

    Parameters
    ----------
    f : function
        The function to find the root of. It should take three parameters:
        sigma, sigma_meas, and D.
    sigma0 : float
        Initial guess for the value of sigma.
    min_sigma : float
        Minimum bound for sigma.
    max_sigma : float
        Maximum bound for sigma.
    tol : float, default: 1e-8
        The tolerance for convergence.
    max_iter : int, default: 100
        The maximum number of iterations.
    *args : tuple
        Additional arguments for the function `f`.

    Returns
    -------
    sigma : float
        The value of sigma that solves the equation, constrained within the
        bounds [min_sigma, max_sigma].
    i : int
        The number of iterations performed. If convergence was reached, this
        is the iteration count. If max iterations were reached, this is equal
        to max_iter.

    Raises
    ------
    ValueError
        If the derivative of the function is near zero, causing a division by
        zero error.

    Notes
    -----
    The method employs Newton's method for root-finding, with safeguards to
    ensure that sigma stays within the bounds [min_sigma, max_sigma]. The
    method terminates when the absolute change in sigma is smaller than the
    specified tolerance `tol` or when the maximum number of iterations is
    reached. If the derivative is too small (near zero), a ValueError is raised.

    """

    sigma = sigma0
    for i in range(max_iter):
        # Evaulate function at current sigma
        f_val = f(sigma, *args)

        # Compute numerical derivative
        delta = tol * sigma
        f_deriv = (f(sigma + delta, *args) - f_val) / delta

        if np.abs(f_deriv) < tol:  # avoid division by zero
            raise ValueError("Derivative near zero, method fails.")

        # Update sigma using Newton-Raphson method.
        delta_sigma = -f_val / f_deriv
        sigma += delta_sigma

        # Apply safeguard to keep sigma within the interval
        # [min_sigma, max_sigma]
        if sigma < min_sigma:
            sigma = min_sigma
        elif sigma > max_sigma:
            sigma = max_sigma

        if np.abs(delta_sigma) < tol:
            return sigma, i  # root found, return solution and iterations

    return sigma, max_iter  # max iterations reached, return last estimate


def complement_gaussian_args(initial_params, fixed_params, fit_params):
    """Complements initial parameters for Gaussian fitting, with
    fixed values.  The end result should be a list of six elements,
    corresponding to 'peak', 'xbar', 'ybar', 'semimajor', 'semiminor'
    and 'theta', in that order

    Parameters
    ----------
    initial_params : np.ndarray
        A Numpy float array containing the initial values for at least one, but
        a maximum of six Gaussian parameters: 'peak', 'xbar', 'ybar',
        'semimajor', 'semiminor' and 'theta' , in that order.
    fixed_params : dict
        A dictionary where the keys are zero, one or more of these parameter
        names: 'peak', 'xbar', 'ybar', 'semimajor', 'semiminor' and 'theta'.
        The values are the fixed values for those parameters. It must complement
        initial_parms, such that the total number of parameters is six:
        len(initial_params) + len(fixed_params) == 6.
    fit_params : tuple
        A tuple of the six Gaussian parameters to be fitted. If a parameter is
        found in `fixed_params`, its value is used from there; otherwise, the
        value is taken from `initial_params`.
        Almost always this should be ('peak', 'xbar', 'ybar', 'semimajor',
        'semiminor', 'theta').
        len(fit_params) == 6.

    Returns
    -------
    gaussian_args : list[float]
        A list of Gaussian fitting arguments, where each value corresponds to
        either a fixed value or an initial value from `initial_params`.
        len(gaussian_args) == 6.

    Example
    -------
    >>> initial_parms = np.array([1.0, 4.0, 5.0, 6.0])
    >>> fixed_parms = {'center_x': 2.5, 'center_y': 3.5}
    >>> fit_parms = ('peak', 'center_x', 'center_y', 'semi-major axis',
    ...               'semi-minor axis', 'position angle')
    >>> complement_gaussian_args(initial_parms, fixed_parms, fit_parms)
    [1.0, 2.5, 3.5, 4.0, 5.0, 6.0]

    """
    paramlist = list(initial_params)
    gaussian_args = []
    for parameter in fit_params:
        if parameter in fixed_params:
            gaussian_args.append(fixed_params[parameter])
        else:
            gaussian_args.append(paramlist.pop(0))

    return gaussian_args



import os
import unittest


def requires_database():
    """Decorator to skip a test if database functionality is
    disabled.

    This function checks the environment variable `TKP_DISABLEDB`. If it is set
    to a truthy value, the test is skipped with an appropriate message.

    Returns
    -------
    function
        A decorator that either skips the test or allows it to run.

    """
    if os.environ.get("TKP_DISABLEDB", False):
        return unittest.skip("Database functionality disabled in configuration")
    return lambda func: func


def requires_data(*args):
    """Decorator to skip a test if required data files are not
    available.

    This function checks for the existence of the specified data files.
    If any of the files do not exist, the test is skipped with an appropriate
    message.

    Parameters
    ----------
    *args : str
        Variable-length argument list of file paths to check.

    Returns
    -------
    function
        A decorator that either skips the test or allows it to run.

    """
    for filename in args:
        if not os.path.exists(filename):
            return unittest.skip("Test data (%s) not available" % filename)
    return lambda func: func


def requires_module(module_name):
    """Decorator to skip a test if a required module is not
    available.

    This function attempts to import the specified module. If the module
    cannot be imported, the test is skipped with an appropriate message.

    Parameters
    ----------
    module_name : str
        The name of the module to check for availability.

    Returns
    -------
    function
        A decorator that either skips the test or allows it to run.

    """
    try:
        __import__(module_name)
    except ImportError:
        return unittest.skip("Required module (%s) not available" % module_name)
    return lambda func: func


def duration(test_duration):
    """Decorator to skip a test if its duration exceeds the maximum
    allowed duration.

    This function checks the environment variable
    ``TKP_MAXTESTDURATION`` to determine the maximum allowed test
    duration. If the test's duration exceeds this value, the test is
    skipped with an appropriate message.

    Parameters
    ----------
    test_duration : float
        The duration of the test in seconds.

    Returns
    -------
    function
        A decorator that either skips the test or allows it to run.

    """
    max_duration = float(os.environ.get("TKP_MAXTESTDURATION", False))
    if max_duration:
        if max_duration < test_duration:
            return unittest.skip(
             "Tests of duration > %s disabled with TKP_MAXTESTDURATION" %
                max_duration)
    return lambda func: func


def requires_test_db_managed():
    """Decorator to disable tests that perform potentially low-level
    database management operations, such as destroying and creating
    databases.

    This decorator checks the environment variables to determine whether
    such tests should be enabled or skipped:

    - If the ``TKP_DBENGINE`` environment variable is set to
      'monetdb', the test is skipped because database management tests
      are not supported for MonetDB and must be tested manually.

    - If the ``TKP_TESTDBMANAGEMENT`` environment variable is set, the
      test is allowed to run.

    - Otherwise, the test is skipped with an appropriate message.

    Returns
    -------
    function
        A decorator that either skips the test or allows it to run.

    """
    if os.environ.get('TKP_DBENGINE', 'postgresql') == 'monetdb':
        return unittest.skip("DB management tests not supported for Monetdb,"
                             "must be tested manually.")

    if os.environ.get("TKP_TESTDBMANAGEMENT", False):
        return lambda func: func
    return unittest.skip("DB management tests disabled, TKP_TESTDBMANAGEMENT"
                         " not set")


def high_ram_requirements():
    """Decorator to disable tests that break Travis due to
    out-of-memory issues.

    This function checks the `TRAVIS` environment variable to determine if the
    tests are running in a Travis CI environment. If so, tests with high RAM
    requirements are skipped with an appropriate message.

    Returns
    -------
    function
        A decorator that either skips the test or allows it to run.

    """
    if os.environ.get("TRAVIS", False):
        return unittest.skip(("High-ram requirement unit-tests disabled on "
                              "Travis"))
    return lambda func: func



from enum import Enum
import pandas as pd


class SourceParams(str, Enum):
    """Enumeration of source parameters that can be measured and stored."""

    PEAK = "peak"
    PEAK_ERR = "peak_err"
    FLUX = "flux"
    FLUX_ERR = "flux_err"
    X = "x"
    X_ERR = "x_err"
    Y = "y"
    Y_ERR = "y_err"
    SMAJ = "smaj"
    SMAJ_ERR = "smaj_err"
    SMIN = "smin"
    SMIN_ERR = "smin_err"
    THETA = "theta"
    THETA_ERR = "theta_err"
    SMAJ_DC = "smaj_dc"
    SMAJ_DC_ERR = "smaj_dc_err"
    SMIN_DC = "smin_dc"
    SMIN_DC_ERR = "smin_dc_err"
    THETA_DC = "theta_dc"
    THETA_DC_ERR = "theta_dc_err"
    RA = "ra"
    RA_ERR = "ra_err"
    DEC = "dec"
    DEC_ERR = "dec_err"
    SMAJ_ASEC = "smaj_asec"
    SMAJ_ASEC_ERR = "smaj_asec_err"
    SMIN_ASEC = "smin_asec"
    SMIN_ASEC_ERR = "smin_asec_err"
    THETA_CELES = "theta_celes"
    THETA_CELES_ERR = "theta_celes_err"
    THETA_DC_CELES = "theta_dc_celes"
    THETA_DC_CELES_ERR = "theta_dc_celes_err"
    ERROR_RADIUS = "error_radius"
    SIG = "sig"
    CHISQ = "chisq"
    REDUCED_CHISQ = "reduced_chisq"

    def describe(self) -> str:
        """Return a description of the source parameter."""
        return _source_params_descriptions[self.value]


_source_params_descriptions = {
    "peak": "Peak spectral brightness of the source (Jy/beam)",
    "peak_err": (
        "1-sigma uncertainty in the peak spectral "
        "brightness of the source (Jy/beam)"
    ),
    "flux": (
        "Flux density of the source, calculated as 'pi * peak "
        "spectral brightness * semi- major axis * semi-minor "
        "axis / beamsize' (Jy)"
    ),
    "flux_err": "1-sigma uncertainty in the flux density (Jy)",
    "x": (
        "x-position (float) of the barycenter of the source, "
        "correponding to the row index of the Numpy array with "
        "image data. After loading a FITS image, the data is "
        "transposed such that x and y are aligned with ds9 viewing, "
        "except for an offset of 1 pixel, since the bottom left "
        "pixel in ds9 has x=y=1"
    ),
    "x_err": (
        "1-sigma uncertainty in the x-position (float) of the "
        "barycenter of the source, corresponding to the row index "
        "of the Numpy array with image data"
    ),
    "y": (
        "y-position (float) of the barycenter of the source, "
        "correponding to the column index of the Numpy array with "
        "image data. After loading a FITS image, the data is "
        "transposed such that x and y are aligned with ds9 viewing, "
        "except for an offset of 1 pixel, since the bottom left "
        "pixel in ds9 has x=y=1"
    ),
    "y_err": (
        "1-sigma uncertainty in the y-position (float) of the "
        "barycenter of the source, corresponding to the column "
        "index of the Numpy array with image data"
    ),
    "smaj": (
        "Semi-major axis of the Gaussian profile, "
        "not deconvolved from the clean beam (pixels)"
    ),
    "smaj_err": (
        "1-sigma uncertainty in the semi-major axis, "
        "not deconvolved from the clean beam (pixels)"
    ),
    "smin": (
        "Semi-minor axis of the Gaussian profile, "
        "not deconvolved from the clean beam (pixels)"
    ),
    "smin_err": (
        "1-sigma uncertainty in the semi-minor axis, "
        "not deconvolved from the clean beam (pixels)"
    ),
    "theta": (
        "Position angle of the major axis of the "
        "Gaussian profile, measured from the positive y-axis "
        " towards the negative x-axis (radians)"
    ),
    "theta_err": (
        "1-sigma uncertainty in the position angle "
        "of the major axis of the Gaussian profile, "
        "measured from the positive y-axis towards the negative x-axis "
        "(radians)"
    ),
    "smaj_dc": (
        "Semi-major axis of the Gaussian profile, "
        "deconvolved from the clean beam (pixels)"
    ),
    "smaj_dc_err": (
        "1-sigma uncertainty in the semi-major axis, "
        "deconvolved from the clean beam (pixels)"
    ),
    "smin_dc": (
        "Semi-minor axis of the Gaussian profile, "
        "deconvolved from the clean beam (pixels)"
    ),
    "smin_dc_err": (
        "1-sigma uncertainty in the semi-minor axis, "
        "deconvolved from the clean beam (pixels)"
    ),
    "theta_dc": (
        "Position angle of the major axis of the "
        "Gaussian profile, deconvolved from the clean beam, "
        "measured from the positive y-axis towards the negative x-axis ("
        "degrees)"
    ),
    "theta_dc_err": (
        "1-sigma uncertainty in the position angle "
        "of the major axis of the Gaussian profile, "
        "deconvolved from the clean beam, measured from the positive y-axis "
        "towards the negative x-axis (degrees)"
    ),
    "ra": "Right ascension of the source (degrees)",
    "ra_err": "1-sigma uncertainty in right ascension (degrees)",
    "dec": "Declination of the source (degrees)",
    "dec_err": "1-sigma uncertainty in declination (degrees)",
    "smaj_asec": (
        "Semi-major axis of the Gaussian profile, "
        "not deconvolved from the clean beam (arcseconds)"
    ),
    "smaj_asec_err": (
        "1-sigma uncertainty in the semi-major axis, "
        "not deconvolved from the clean beam "
        "(arcseconds)"
    ),
    "smin_asec": (
        "Semi-minor axis of the Gaussian profile, "
        "not deconvolved from the clean beam (arcsecond)"
    ),
    "smin_asec_err": (
        "1-sigma uncertainty in the semi-minor axis, "
        "not deconvolved from the clean beam "
        "(arcseconds)"
    ),
    "theta_celes": (
        "Position angle of the major axis of the "
        "Gaussian profile, measured east from local north "
        "(degrees)"
    ),
    "theta_celes_err": (
        "1-sigma uncertainty in the position angle "
        "of the major axis of the Gaussian profile, "
        "measured east from local north (degrees)"
    ),
    "theta_dc_celes": (
        "Position angle of the major axis of the "
        "Gaussian profile, deconvolved from the clean beam, "
        "measured east from local north (degrees)"
    ),
    "theta_dc_celes_err": (
        "1-sigma uncertainty in the position angle "
        "of the major axis of the Gaussian profile, "
        "deconvolved from the clean beam, measured east from local north "
        "(degrees)"
    ),
    "error_radius": (
        "The absolute angular error on the position of the source "
        "(arcseconds). This is a pessimistic estimate, because we try "
        "all possible combinations of the x and y errors, and take the "
        "maximum for the four combinations."
    ),
    "sig": (
        "The significance of a detection (float) is defined as "
        "the maximum signal-to-noise ratio across the island. "
        "Often this will be the ratio of the maximum pixel value "
        "of the source divided by the noise at that position."
    ),
    "chisq": (
        "The chi-squared value of the Gaussian model relative to "
        "the data (float). Can be a Gaussian model derived from a fit "
        "or from moments. See the measuring.goodness_of_fit docstring "
        "for some important notes."
    ),
    "reduced_chisq": (
        "The reduced chi-squared value of the Gaussian "
        "model relative to the data (float). Can be a "
        "Gaussian model derived from a fit or from "
        "moments. See the measuring.goodness_of_fit "
        "docstring for some important notes."
    ),
}

# Ensure that all source parameters have a description
assert all(p.value in _source_params_descriptions for p in SourceParams)

# This should render source parameter descriptions in Sphinx/RTD.
for member in SourceParams:
    member.__doc__ = SourceParams.describe(member)

# Set default set of source parameters to store in a file, e.g. a .csv file.
_file_fields = [
    "PEAK",
    "PEAK_ERR",
    "FLUX",
    "FLUX_ERR",
    "X",
    "Y",
    "RA",
    "RA_ERR",
    "DEC",
    "DEC_ERR",
    "SMAJ_ASEC",
    "SMAJ_ASEC_ERR",
    "SMIN_ASEC",
    "SMIN_ASEC_ERR",
    "THETA_CELES",
    "THETA_CELES_ERR",
    "SIG",
    "REDUCED_CHISQ",
]


def make_measurements_dataframe(
    moments_of_sources,
    sky_barycenters,
    ra_errors,
    dec_errors,
    smaj_asec,
    errsmaj_asec,
    smin_asec,
    errsmin_asec,
    theta_celes_values,
    theta_celes_errors,
    theta_dc_celes_values,
    theta_dc_celes_errors,
    error_radii,
    sig,
    chisq,
    reduced_chisq,
):
    """
    Create a Pandas DataFrame with parameters related to detected sources
    from a subset of the tuple of Numpy ndarrays returned by the
    `extract.source_measurements_vectorised` function.
    """
    columns = {
        SourceParams.PEAK: moments_of_sources[:, 0, 0],
        SourceParams.PEAK_ERR: moments_of_sources[:, 1, 0],
        SourceParams.FLUX: moments_of_sources[:, 0, 1],
        SourceParams.FLUX_ERR: moments_of_sources[:, 1, 1],
        SourceParams.X: moments_of_sources[:, 0, 2],
        SourceParams.X_ERR: moments_of_sources[:, 1, 2],
        SourceParams.Y: moments_of_sources[:, 0, 3],
        SourceParams.Y_ERR: moments_of_sources[:, 1, 3],
        SourceParams.SMAJ: moments_of_sources[:, 0, 4],
        SourceParams.SMAJ_ERR: moments_of_sources[:, 1, 4],
        SourceParams.SMIN: moments_of_sources[:, 0, 5],
        SourceParams.SMIN_ERR: moments_of_sources[:, 1, 5],
        SourceParams.THETA: moments_of_sources[:, 0, 6],
        SourceParams.THETA_ERR: moments_of_sources[:, 1, 6],
        SourceParams.SMAJ_DC: moments_of_sources[:, 0, 7],
        SourceParams.SMAJ_DC_ERR: moments_of_sources[:, 1, 7],
        SourceParams.SMIN_DC: moments_of_sources[:, 0, 8],
        SourceParams.SMIN_DC_ERR: moments_of_sources[:, 1, 8],
        SourceParams.THETA_DC: moments_of_sources[:, 0, 9],
        SourceParams.THETA_DC_ERR: moments_of_sources[:, 1, 9],
        SourceParams.RA: sky_barycenters[:, 0],
        SourceParams.RA_ERR: ra_errors.ravel(),
        SourceParams.DEC: sky_barycenters[:, 1],
        SourceParams.DEC_ERR: dec_errors.ravel(),
        SourceParams.SMAJ_ASEC: smaj_asec,
        SourceParams.SMAJ_ASEC_ERR: errsmaj_asec,
        SourceParams.SMIN_ASEC: smin_asec,
        SourceParams.SMIN_ASEC_ERR: errsmin_asec,
        SourceParams.THETA_CELES: theta_celes_values,
        SourceParams.THETA_CELES_ERR: theta_celes_errors.ravel(),
        SourceParams.THETA_DC_CELES: theta_dc_celes_values.ravel(),
        SourceParams.THETA_DC_CELES_ERR: theta_dc_celes_errors.ravel(),
        SourceParams.ERROR_RADIUS: error_radii.ravel(),
        SourceParams.SIG: sig,
        SourceParams.CHISQ: chisq,
        SourceParams.REDUCED_CHISQ: reduced_chisq,
    }

    return pd.DataFrame(columns)


def describe_dataframe_columns(df: pd.DataFrame) -> dict[str, str]:
    return {
        col: _source_params_descriptions[col]
        for col in df.columns
        if col in _source_params_descriptions
    }
